/*
 * $Id: bgp_attr.c,v 1.20 1998/07/17 20:37:47 masaki Exp $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>

#include <mrt.h>
#include <array.h>
#include <New.h>
#include <hash.h>
#include <io.h>
#include <interface.h>
#include <bgp.h>

/*
 * BGP4+ code was originally written 
 *     by Francis Dupont <Francis.Dupont@inria.fr>.
 */


static int num_active_bgp_attr = 0;

static char *s_origins[] = {"IGP", "EGP", "INCOMPLETE", "AGGREGATE"};
const char * origin2string (int origin) { return (s_origins[origin]); }


/* 
 * Process an MRT BGP update message. Fill in pp_attr
 * and ll_ann_prefixes and ll_with_prefixes.
 */
int
bgp_process_update_msg (enum MRT_MSG_TYPES type, u_char * value, int length,
			gateway_t ** gateway_to, bgp_attr_t ** pp_attr,
			LINKED_LIST ** ll_with_prefixes,
			LINKED_LIST ** ll_ann_prefixes)
{
#ifdef HAVE_IPV6
    struct in6_addr addr;
#else
    u_long addr;
#endif /* HAVE_IPV6 */
    int as, ret;
    prefix_t *prefix;
    gateway_t *gateway_from;
    int plen = 4;
    int family = AF_INET;
    u_char *cp = value;
    int bgp4plus = DEFAULT_BGP4PLUS_VERSION;
    bgp_peer_t *peer;
    char str[MAXLINE];

#ifdef HAVE_IPV6
    if (type == MSG_PROTOCOL_BGP4PLUS) {
	plen = 16;
	family = AF_INET6;
    }
    if (type == MSG_PROTOCOL_BGP4PLUS_01) {
	plen = 16;
	family = AF_INET6;
	bgp4plus = 1;
    }
#endif /* HAVE_IPV6 */

    /* create a pseudo peer */
    peer = New (bgp_peer_t);
    peer->trace = trace_copy (MRT->trace);
    if (bgp4plus) {
        BIT_SET (peer->options, BGP_BGP4PLUS_01);
    }

    /* from */
    BGP_GET_SHORT (as, cp);
    memcpy (&addr, cp, plen);
    cp += plen;
    prefix = New_Prefix (family, &addr, plen * 8);
    gateway_from = add_gateway (prefix, as, NULL);

#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6)
        sprintf (str, "BGP4+ %s", prefix_toa (prefix));
    else
#endif /* HAVE_IPV6 */
        sprintf (str, "BGP %s", prefix_toa (prefix));
    set_trace (peer->trace, TRACE_PREPEND_STRING, str, 0);
    Deref_Prefix (prefix);

    /* to */
    BGP_GET_SHORT (as, cp);
    memcpy (&addr, cp, plen);
    cp += plen;
    prefix = New_Prefix (family, &addr, plen * 8);
    *gateway_to = add_gateway (prefix, as, NULL);
    Deref_Prefix (prefix);

    length -= (cp - value);

    if ((ret = bgp_process_update_packet (cp, length, peer)) < 0) {
	/* code and subcode available */
    }
    else {
        /* in case of withdraw only */
        if (peer->attr == NULL)
	    peer->attr = bgp_new_attr (PROTO_BGP);
        peer->attr->gateway = gateway_from;
    }

    /* setting return values */
    *pp_attr = peer->attr;
    *ll_with_prefixes = peer->ll_withdraw;
    *ll_ann_prefixes = peer->ll_announce;

    Destroy_Trace (peer->trace);
    Delete (peer);

    return (ret);
}


/*
 * Given a pointer (and size) of BGP packet in memory, 
 * return linked lists of withdrawn prefixes, announced prefixes
 * and a BGP attribute structure
 */
int
bgp_process_update_packet (u_char * cp, int length, bgp_peer_t *peer)
{
    u_char *cpend = cp + length;
    int withlen;
    int attrlen;
    u_char *attrp;

    assert (peer);
    peer->ll_withdraw = NULL;
    peer->ll_announce = NULL;
    peer->attr = NULL;

    BGP_GET_V4UPDATE_UNREACH (withlen, cp);

    if (withlen + cp + BGP_ATTR_SIZE_LEN > cpend) {
	trace (TR_ERROR, peer->trace, "bad withdraw length %d\n", withlen);
	peer->subcode = BGP_ERRUPD_ATTRLIST; /* XXX */
	goto error;
    }

    if (withlen > 0) {

	u_char *cpend = cp + withlen;

	if (peer->ll_withdraw == NULL) {
	    peer->ll_withdraw = LL_Create (LL_DestroyFunction, Deref_Prefix, 0);
	}

        trace (TR_PACKET, peer->trace, "recv withdraw:\n");

        while (cp < cpend) {
            int bitlen;
            u_char dest[4];
            prefix_t *prefix;

	    BGP_GET_BITCOUNT (bitlen, cp);
	    if (bitlen < 0 || bitlen > 32) {
	        trace (TR_ERROR, peer->trace, "bad bitlen %d (withdraw)\n", 
		       bitlen);
		peer->subcode = BGP_ERRUPD_BADNET;
		goto error;
	    }
	    BGP_GET_PREFIX (bitlen, dest, cp);
	    prefix = New_Prefix (AF_INET, dest, bitlen);
	    LL_Add (peer->ll_withdraw, prefix);
	    trace (TR_PACKET, peer->trace, "  %s\n", prefix_toax (prefix));
        }
        if (cp != cpend) {
	    trace (TR_ERROR, peer->trace, "bad byte count %d (withdraw)\n", 
		   cp - cpend);
	    peer->subcode = BGP_ERRUPD_BADNET;
	    goto error;
        }
    }

    BGP_GET_UPDATE (attrlen, attrp, cp);

    if (attrlen + attrp > cpend) {
	trace (TR_ERROR, peer->trace, "bad attribute length %d\n", attrlen);
	peer->subcode = BGP_ERRUPD_ATTRLIST;
	goto error;
    }
    if (attrlen) {
        if (bgp_munge_attributes (attrlen, attrp, peer) == NULL &&
	    peer->code > 0)
	    return (-1);
    }

    if (cp >= cpend)
	return (0);

    trace (TR_PACKET, peer->trace, "recv announce:\n");
    while (cp < cpend) {
	int bitlen;
        u_char dest[4];
	prefix_t *prefix;

	BGP_GET_BITCOUNT (bitlen, cp);
	if (bitlen < 0 || bitlen > 32) {
	    trace (TR_ERROR, peer->trace, "bad bitlen %d (announce)\n", bitlen);
	    peer->subcode = BGP_ERRUPD_BADNET;
	    goto error;
	}
	BGP_GET_PREFIX (bitlen, dest, cp);
	prefix = New_Prefix (AF_INET, dest, bitlen);

	if (peer->ll_announce == NULL) {
	    peer->ll_announce = LL_Create (LL_DestroyFunction,
				      Deref_Prefix, 0);
	}

	LL_Add (peer->ll_announce, prefix);
	trace (TR_PACKET, peer->trace, "  %s\n", prefix_toax (prefix));
    }
    if (cp != cpend) {
	trace (TR_ERROR, peer->trace, "bad byte count %d (announce)\n", 
	      cp - cpend);
	peer->subcode = BGP_ERRUPD_ATTRLIST;
error:
	peer->code = BGP_ERR_UPDATE;
	if (peer->ll_withdraw) {
	    LL_Destroy (peer->ll_withdraw);
	    peer->ll_withdraw = NULL;
	}
	if (peer->ll_announce) {
	    LL_Destroy (peer->ll_announce);
	    peer->ll_announce = NULL;
	}
	if (peer->attr) {
	    bgp_deref_attr (peer->attr);
	    peer->attr = NULL;
	}
	return (-1);
    }
    return (0);
}


#ifdef HAVE_IPV6

#define BGP_DEFAULT_IPV6_FAMILY 2
#define BGP_DEFAULT_IPV6_SUBFAMILY 1

int
bgp_get_announced_ipv6_prefixes (u_char * cp, u_char * cpend, bgp_peer_t *peer)
{
    u_char dest[16];
    u_short family, plen;
    u_char subfamily, nhalen, nsnpa;
    int i;
    char tmp6[INET6_ADDRSTRLEN];
    u_char lladdr[16];
    int autoset = 0;

    int bgp4plus = (BIT_TEST (peer->options, BGP_BGP4PLUS_01)? 1: 0);

    if (cpend - cp < 5) {
	trace (TR_ERROR, peer->trace,
	       "recv invalid length %d detected (announce)\n", cpend - cp);
	peer->code = BGP_ERR_UPDATE;
	peer->subcode = BGP_ERRUPD_OPTATTR;
	return (-1);
    }

    BGP_GET_SHORT (family, cp);
    BGP_GET_BYTE (subfamily, cp);
    BGP_GET_BYTE (nhalen, cp);

    trace (TR_PACKET, peer->trace,
	   "announce family %d subfamily %d nhalen %d\n",
	   family, subfamily, nhalen);

    if (family != BGP_DEFAULT_IPV6_FAMILY) {
	trace (TR_ERROR, peer->trace,
	       "recv announce unrecognize family %d\n", family);
	peer->code = BGP_ERR_UPDATE;
	peer->subcode = BGP_ERRUPD_OPTATTR;
	return (-1);
    }
    if (subfamily != BGP_DEFAULT_IPV6_SUBFAMILY) {
	trace (TR_ERROR, peer->trace,
	       "recv announce unrecognize subfamily %d\n", subfamily);
	peer->code = BGP_ERR_UPDATE;
	peer->subcode = BGP_ERRUPD_OPTATTR;
	return (-1);
    }
    if (nhalen != 16 && nhalen != 32) {
	trace (TR_ERROR, peer->trace, 
	       "recv announce unrecognize nhalen %d\n", nhalen);
	peer->code = BGP_ERR_UPDATE;
	peer->subcode = BGP_ERRUPD_OPTATTR;
	return (-1);
    }

    for (i = 0; i < 16; i++)
	BGP_GET_BYTE (dest[i], cp);

    trace (TR_PACKET, peer->trace, "  NEXT_HOP: %s\n",
	   inet_ntop (AF_INET6, dest, tmp6, sizeof tmp6));

    if (nhalen == 32) {
	trace (TR_PACKET, peer->trace, "  NEXT_HOP: %s\n",
	       inet_ntop (AF_INET6, cp, tmp6, sizeof tmp6));
	memcpy (lladdr, cp, 16);
	cp += 16;
    }

    BGP_GET_BYTE (nsnpa, cp);
    if (nsnpa > 0) {
	trace (TR_PACKET, peer->trace,
	       "recv announce Number of SNPAs = %d\n", nsnpa);
	/* skip SNPAs */
	for (i = 0; i < (u_int) nsnpa; i++) {
	    BGP_GET_BYTE (nhalen, cp);
	    cp += (nhalen + 1) >> 1;
	}
    }

    BGP_GET_SHORT (plen, cp);
    if (cpend == cp + plen && bgp4plus != 0) {
	trace (TR_ERROR, peer->trace,
	       "recv announce draft version should be 0\n");
	bgp4plus = 0;
	if (BIT_TEST (peer->options, BGP_BGP4PLUS_AUTO)) {
    	    autoset++;
	}
	else {
	    /* send notification */
	}
    }
    if (cpend != cp + plen && bgp4plus == 0) {
	trace (TR_ERROR, peer->trace,
	       "recv announce draft version should be 1\n");
	bgp4plus = 1;
	if (BIT_TEST (peer->options, BGP_BGP4PLUS_AUTO)) {
    	    autoset++;
	}
	else {
	    /* send notification */
	}
    }
    cp -= 2;

    if (bgp4plus == 0) {
	BGP_GET_SHORT (plen, cp);
	trace (TR_PACKET, peer->trace,
	       "recv announce draft 0 len %d\n", plen);
	if (cp + plen != cpend) {
	    trace (TR_ERROR, peer->trace,
		   "recv announce left %d bytes at the end\n",
		   cpend - (cp + plen));
	    peer->code = BGP_ERR_UPDATE;
	    peer->subcode = BGP_ERRUPD_OPTATTR;
	    return (-1);
	}
	if (plen <= 0)
	    return (0);
	cpend = cp + plen;
    }

    assert (peer->attr);
    if (peer->attr->nexthop) {
	assert (peer->attr->nexthop->family == AF_INET);
	peer->attr->nexthop4 = peer->attr->nexthop;
	/* save ipv4 nexthop */
    }
    /* XXX do we need to keep v4/v6 nexthop ? */
    peer->attr->nexthop = New_Prefix (AF_INET6, dest, 128);
    /* XXX old version of MRT puts 0 for link local */
    if (nhalen == 32 && !IN6_IS_ADDR_UNSPECIFIED ((struct in6_addr *)lladdr)) {
	peer->attr->link_local = New_Prefix (AF_INET6, lladdr, 128);
    }
    /* XXX so, PA4_TYPE_NEXTHOP changes its meaning */
    BGP4_BIT_SET (peer->attr->attribs, PA4_TYPE_NEXTHOP);

    trace (TR_PACKET, peer->trace, "recv announce:\n");
    while (cp < cpend) {
        u_int bitlen;
	prefix_t *prefix;

	BGP_GET_BITCOUNT (bitlen, cp);
	if (bitlen < 0 || bitlen > 128) {
	    trace (TR_ERROR, peer->trace, "bad bitlen %d (announce)\n", bitlen);
	    peer->code = BGP_ERR_UPDATE;
	    peer->subcode = BGP_ERRUPD_BADNET;
	    return (-1);
	}
	BGP_GET_PREFIX6 (bitlen, dest, cp);

	prefix = New_Prefix (AF_INET6, dest, bitlen);
	trace (TR_PACKET, peer->trace, "  %s\n", prefix_toax (prefix));
	if (peer->ll_announce == NULL)
	    peer->ll_announce = LL_Create (LL_DestroyFunction, Deref_Prefix, 0);
	LL_Add (peer->ll_announce, prefix);
    }
    if (cp != cpend) {
	trace (TR_ERROR, peer->trace, "bad byte count %d (announce)\n", 
	      cp - cpend);
	peer->code = BGP_ERR_UPDATE;
	peer->subcode = BGP_ERRUPD_ATTRLIST;
	return (-1);
    }

    /*
     * Only if the packet was processed, fix the version automatically
     */
    if (autoset) {
	if (bgp4plus)
    	    BIT_SET (peer->options, BGP_BGP4PLUS_01);
	else
    	    BIT_RESET (peer->options, BGP_BGP4PLUS_01);
    }
    return (1);
}


int
bgp_get_withdrawn_ipv6_prefixes (u_char * cp, u_char * cpend, bgp_peer_t *peer)
{
    u_short family, plen;
    u_char subfamily;
    u_char dest[16];
    int autoset = 0;

    int bgp4plus = (BIT_TEST (peer->options, BGP_BGP4PLUS_01)? 1: 0);

    if (cpend - cp < 3) {
	trace (TR_ERROR, peer->trace,
	       "recv invalid length %d detected (withdraw)\n", cpend - cp);
	peer->code = BGP_ERR_UPDATE;
	peer->subcode = BGP_ERRUPD_OPTATTR;
	return (-1);
    }

    BGP_GET_SHORT (family, cp);
    BGP_GET_BYTE (subfamily, cp);
    trace (TR_PACKET, peer->trace, "recv withdraw family %d subfamily %d\n",
	   family, subfamily);

    if (family != BGP_DEFAULT_IPV6_FAMILY) {
	trace (TR_ERROR, peer->trace,
	       "recv withdraw unrecognize family %d\n", family);
	peer->code = BGP_ERR_UPDATE;
	peer->subcode = BGP_ERRUPD_OPTATTR;
	return (-1);
    }
    if (subfamily != BGP_DEFAULT_IPV6_SUBFAMILY) {
	trace (TR_ERROR, peer->trace,
	       "recv withdraw unrecognize subfamily %d\n", subfamily);
	peer->code = BGP_ERR_UPDATE;
	peer->subcode = BGP_ERRUPD_OPTATTR;
	return (-1);
    }

    BGP_GET_SHORT (plen, cp);
    if (cpend == cp + plen && bgp4plus != 0) {
	trace (TR_ERROR, peer->trace,
	       "recv withdraw draft version should be 0\n");
	bgp4plus = 0;
	if (BIT_TEST (peer->options, BGP_BGP4PLUS_AUTO)) {
    	    autoset++;
	}
	else {
	    /* send notification */
	}
    }
    if (cpend != cp + plen && bgp4plus == 0) {
	trace (TR_ERROR, peer->trace,
	       "recv withdraw draft version should be 1\n");
	bgp4plus = 1;
	if (BIT_TEST (peer->options, BGP_BGP4PLUS_AUTO)) {
    	    autoset++;
	}
	else {
	    /* send notification */
	}
    }
    cp -= 2;

    if (bgp4plus == 0) {
	BGP_GET_SHORT (plen, cp);
	trace (TR_PACKET, peer->trace,
	       "recv withdraw draft 0 len %d\n", plen);
	if (cpend != cp + plen) {
	    trace (TR_ERROR, peer->trace,
		   "recv withdraw left %d bytes at the end\n",
		   cpend - (cp + plen));
	    peer->code = BGP_ERR_UPDATE;
	    peer->subcode = BGP_ERRUPD_OPTATTR;
	    return (-1);
	}
	cpend = cp + plen;

	if (plen <= 0)
	    return (0);
    }

    trace (TR_PACKET, peer->trace, "recv withdraw:\n");
    while (cp < cpend) {
        u_int bitlen;
	prefix_t *prefix;

	BGP_GET_BITCOUNT (bitlen, cp);
	if (bitlen < 0 || bitlen > 128) {
	    trace (TR_ERROR, peer->trace, "bad bitlen %d (withdraw)\n", bitlen);
	    peer->code = BGP_ERR_UPDATE;
	    peer->subcode = BGP_ERRUPD_BADNET;
	    return (-1);
	}
	BGP_GET_PREFIX6 (bitlen, dest, cp);

	prefix = New_Prefix (AF_INET6, dest, bitlen);
	trace (TR_PACKET, peer->trace, "  %s\n", prefix_toax (prefix));
	if (peer->ll_withdraw == NULL)
	    peer->ll_withdraw = LL_Create (LL_DestroyFunction, Deref_Prefix, 0);
	LL_Add (peer->ll_withdraw, prefix);
    }

    if (cp != cpend) {
	trace (TR_ERROR, peer->trace, "bad byte count %d (withdraw)\n", 
	      cp - cpend);
	peer->code = BGP_ERR_UPDATE;
	peer->subcode = BGP_ERRUPD_ATTRLIST;
	return (-1);
    }

    /*
     * Only if the packet was processed, fix the version automatically
     */
    if (autoset) {
	if (bgp4plus)
    	    BIT_SET (peer->options, BGP_BGP4PLUS_01);
	else
    	    BIT_RESET (peer->options, BGP_BGP4PLUS_01);
    }
    return (1);
}
#endif /* HAVE_IPV6 */


/* 
 * Given a length and byte string from a BGP4 packet, return origin,
 * linked list of as_path_segments, and other attributes
 */
bgp_attr_t *
bgp_munge_attributes (int attrlen, u_char * cp, bgp_peer_t *peer)
{
    u_char *endp;
    u_int flags;
    u_int code;
    u_char seen[256];
    int len = 0;
    u_long nexthop;
    u_long value;
    u_char *at;

    assert (peer);
    assert (peer->attr == NULL);
    peer->attr = bgp_new_attr (PROTO_BGP);
    memset (seen, 0, sizeof seen);
    endp = cp + attrlen;

    trace (TR_PACKET, peer->trace, "recv attribute:\n");
    while (cp < endp) {

	if (endp - cp < 3) {
	    trace (TR_ERROR, peer->trace, 
		   "bad remaining length %d\n", endp - cp);
	    peer->subcode = BGP_ERRUPD_ATTRLIST;
	    goto error;
	}
	GET_PATH_ATTR (flags, code, len, cp);
	at = cp;

#ifdef notdef
	trace (TR_PACKET, peer->trace,
	       "recv flags = %x, code %d, len = %d, cp = %x, endp = %x\n",
	       flags, code, len, cp, endp);
#endif

	if (seen[code]) {
	    /* duplicate code */
	    trace (TR_ERROR, peer->trace, 
		   "duplicate attribute: flags 0x%x code %d len %d\n", 
		    flags, code, len);
	    peer->subcode = BGP_ERRUPD_ATTRLIST;
	    goto error;
	}
	else {
	    seen[code]++;
	}

	if (cp + len > endp) {
	    trace (TR_ERROR, peer->trace, "bad attribute length %d\n", len);
	    peer->subcode = BGP_ERRUPD_LENGTH;
	    goto error;
	}

	/* check the case that should not happen */
	if (!BIT_TEST(flags, PA_FLAG_OPT)) {
            if (!BIT_TEST(flags, PA_FLAG_TRANS) || 
		 BIT_TEST(flags, PA_FLAG_PARTIAL)) {
	        peer->subcode = BGP_ERRUPD_FLAGS;
		goto error;
            }
        }

	switch (code) {
	case PA4_TYPE_INVALID:
	default:
	    break;
	case PA4_TYPE_ORIGIN:
	case PA4_TYPE_ASPATH:
	case PA4_TYPE_NEXTHOP:
	case PA4_TYPE_LOCALPREF:
	case PA4_TYPE_ATOMICAGG:
	    /* Well-known Transitive */
	    if ((flags & PA_FLAG_ALL) != PA_FLAG_TRANS) {
	        trace (TR_ERROR, peer->trace, "bad flags 0x%x (code %d)\n", 
		      flags, code);
	        peer->subcode = BGP_ERRUPD_FLAGS;
		goto error;
	    }
	    break;
	case PA4_TYPE_METRIC:	/* multiexit */
#ifdef HAVE_IPV6
	case PA4_TYPE_MPUNRNLRI:
	case PA4_TYPE_MPREACHNLRI:
#endif /* HAVE_IPV6 */
	    /* Optional Non-Transitive */
	    if ((flags & PA_FLAG_ALL) != PA_FLAG_OPT) {
	        trace (TR_ERROR, peer->trace, "bad flags 0x%x (code %d)\n", 
		      flags, code);
	        peer->subcode = BGP_ERRUPD_FLAGS;
		goto error;
	    }
	    break;
	case PA4_TYPE_COMMUNITY:
	case PA4_TYPE_AGGREGATOR:
	case PA4_TYPE_DPA:
	    /* Optional Transitive */
	    if ((flags & PA_FLAG_ALL) != (PA_FLAG_OPT | PA_FLAG_TRANS)) {
	        trace (TR_ERROR, peer->trace, "bad flags 0x%x (code %d)\n", 
		      flags, code);
	        peer->subcode = BGP_ERRUPD_FLAGS;
		goto error;
	    }
	    break;
	}

	switch (code) {
	case PA4_TYPE_INVALID:
	    trace (TR_ERROR, peer->trace, "code %d is invalid\n");
	    peer->subcode = BGP_ERRUPD_ATTRLIST;
	    goto error;
	case PA4_TYPE_ORIGIN:
	    if (len != 1) {
	        trace (TR_ERROR, peer->trace, "bad length %d (origin)\n", len);
	        peer->subcode = BGP_ERRUPD_LENGTH;
		goto error;
	    }
	    BGP_GET_BYTE (peer->attr->origin, cp);
	    if (/* peer->attr->origin < 0 || */ peer->attr->origin > 2) {
	        trace (TR_ERROR, peer->trace, "bad prigin code %d\n", 
		       peer->attr->origin);
	        peer->subcode = BGP_ERRUPD_ORIGIN;
		goto error;
	    }
	    trace (TR_PACKET, peer->trace, "  ORIGIN: %s\n",
		   origin2string (peer->attr->origin));
	    BGP4_BIT_SET (peer->attr->attribs, code);
	    break;
	case PA4_TYPE_ASPATH:
	    if (len == 0) {
		/* NULL aspath for iBGP */
	        peer->attr->aspath = NULL;
	        peer->attr->home_AS = 0;
	        trace (TR_PACKET, peer->trace, "  ASPATH: %s\n",
		       aspath_toa (peer->attr->aspath));
	        BGP4_BIT_SET (peer->attr->attribs, code);
	    }
	    else {
	        if ((peer->attr->aspath = munge_aspath (len, cp)) == NULL) {
		    peer->subcode = BGP_ERRUPD_ASPATH;
	            goto error;
		}
		peer->attr->aspath = aspath_reduce (peer->attr->aspath);
	        peer->attr->home_AS = bgp_get_home_AS (peer->attr->aspath);
	        trace (TR_PACKET, peer->trace, "  ASPATH: %s\n",
		       aspath_toa (peer->attr->aspath));
	        cp += len;
	        BGP4_BIT_SET (peer->attr->attribs, code);
	    }
	    break;
	case PA4_TYPE_NEXTHOP:
	    if (len != 4) {
	        trace (TR_ERROR, peer->trace, "bad length %d (nexthop)\n", len);
		peer->subcode = BGP_ERRUPD_LENGTH;
	        goto error;
	    }
	    BGP_GET_NETLONG (nexthop, cp);
	    peer->attr->nexthop = New_Prefix (AF_INET, (u_char *) & nexthop, 
					      32);
	    trace (TR_PACKET, peer->trace, "  NEXT_HOP: %s\n",
		   prefix_toa (peer->attr->nexthop));
	    BGP4_BIT_SET (peer->attr->attribs, code);
	    break;
	case PA4_TYPE_METRIC:	/* multiexit */
	    if (len != 4) {
	        trace (TR_ERROR, peer->trace, "bad length %d (metric)\n", len);
		peer->subcode = BGP_ERRUPD_LENGTH;
	        goto error;
	    }
	    BGP_GET_LONG (peer->attr->multiexit, cp);
	    trace (TR_PACKET, peer->trace, "  MULTI_EXIT_DISC: %d\n", 
		   peer->attr->multiexit);
	    BGP4_BIT_SET (peer->attr->attribs, code);
	    break;
	case PA4_TYPE_LOCALPREF:
	    if (len != 4) {
	        trace (TR_ERROR, peer->trace, "bad length %d (localpref)\n", 
		       len);
		peer->subcode = BGP_ERRUPD_LENGTH;
	        goto error;
	    }
	    BGP_GET_LONG (peer->attr->local_pref, cp);
	    trace (TR_PACKET, peer->trace, "  LOCAL_PREF: %d\n", 
		   peer->attr->local_pref);
	    BGP4_BIT_SET (peer->attr->attribs, code);
	    break;
	case PA4_TYPE_ATOMICAGG:
	    if (len != 0) {
	        trace (TR_ERROR, peer->trace, "bad length %d (atomicagg)\n", 
		       len);
		peer->subcode = BGP_ERRUPD_LENGTH;
	        goto error;
	    }
	    trace (TR_PACKET, peer->trace, "  ATOMIC_AGGREGATE\n");
	    BGP4_BIT_SET (peer->attr->attribs, code);
	    break;
	case PA4_TYPE_AGGREGATOR:
	    if (len != 6) {
	        trace (TR_ERROR, peer->trace, "bad length %d (aggregator)\n", 
		       len);
		peer->subcode = BGP_ERRUPD_LENGTH;
	        goto error;
	    }
	    BGP_GET_SHORT (peer->attr->aggregator.as, cp);
	    BGP_GET_NETLONG (value, cp);
	    peer->attr->aggregator.prefix = 
			New_Prefix (AF_INET, (u_char *) & value, 32);
	    trace (TR_PACKET, peer->trace, "  AGGREGATOR: AS%d %s\n",
		   peer->attr->aggregator.as, 
		   prefix_toa (peer->attr->aggregator.prefix));
	    BGP4_BIT_SET (peer->attr->attribs, code);
	    break;
	case PA4_TYPE_COMMUNITY:
	    if (len > 0) {
	        if ((peer->attr->community = munge_community (len, cp)) 
			== NULL) {
		    peer->subcode = BGP_ERRUPD_OPTATTR;
	            goto error;
		}
	        trace (TR_PACKET, peer->trace, "  COMMUNITY: %s\n",
		       community_toa (peer->attr->community));
	        cp += len;
	        BGP4_BIT_SET (peer->attr->attribs, code);
	    }
	    break;
	case PA4_TYPE_DPA:
	    if (len != 6) {
	        trace (TR_ERROR, peer->trace, "bad length %d (dpa)\n", 
		       len);
		peer->subcode = BGP_ERRUPD_LENGTH;
	        goto error;
	    }
	    BGP_GET_SHORT (peer->attr->dpa.as, cp);
	    BGP_GET_LONG (peer->attr->dpa.value, cp);
	    trace (TR_PACKET, peer->trace, "  DPA: AS%d %d\n",
		   peer->attr->dpa.as, peer->attr->dpa.value);
	    BGP4_BIT_SET (peer->attr->attribs, code);
	    break;
#ifdef HAVE_IPV6
	case PA4_TYPE_MPUNRNLRI:
	    if (len > 0) {
	        if (bgp_get_withdrawn_ipv6_prefixes (cp, cp + len, peer) < 0) {
		    /* the function sets codes */
                    bgp_deref_attr (peer->attr);
	            peer->attr = NULL;
        	    peer->code = peer->subcode = 0;
		    return (NULL);
	        }
	        cp += len;
	        /* don't need to set attr->attribs with this code */
	    }
	    break;
	case PA4_TYPE_MPREACHNLRI:
	    if (len > 0) {
	        if (bgp_get_announced_ipv6_prefixes (cp, cp + len, peer) < 0) {
		    /* the function sets codes */
                    bgp_deref_attr (peer->attr);
	            peer->attr = NULL;
        	    peer->code = peer->subcode = 0;
		    return (NULL);
	        }
	        cp += len;
	        /* don't need to set attr->attribs with this code */
	    }
	    break;
#endif /* HAVE_IPV6 */
	default:
	    trace (TR_WARN, peer->trace, "recv %s %s %s attribute: "
		   "flags 0x%x code %d len %d\n", 
		BIT_TEST (flags, PA_FLAG_OPT)? "optional": "mandatory",
		BIT_TEST (flags, PA_FLAG_TRANS)? "transitive": "non-transitive",
		BIT_TEST (flags, PA_FLAG_PARTIAL)? "partial": "non-partial", 
		flags, code, len);
	    /* XXX MRT doesn't pass unknown transitive attributes */
	    cp += len;

	    if (!BIT_TEST (flags, PA_FLAG_OPT)) {
	        trace (TR_ERROR, peer->trace, "unknown well-known attribute: "
		   "flags 0x%x code %d len %d\n", flags, code, len);
		peer->subcode = BGP_ERRUPD_UNKNOWN;
	        goto error;
	    }
	    break;
	}

	if (at + len != cp) {
	    trace (TR_ERROR, peer->trace, 
		   "bad attribute length %d (should be %d)\n", len, cp - at);
	    peer->subcode = BGP_ERRUPD_LENGTH;
	    goto error;
	}
    }

    if (cp != endp) {
	trace (TR_ERROR, peer->trace, "recv byte count %d\n", cp - endp);
	peer->subcode = BGP_ERRUPD_LENGTH;
error:
	peer->code = BGP_ERR_UPDATE;
        bgp_deref_attr (peer->attr);
	peer->attr = NULL;
        peer->code = peer->subcode = 0;
	return (NULL);
    }

    if (peer->attr->attribs == 0) {
	/* No attributes. Probably only IPv6 stuff */
        bgp_deref_attr (peer->attr);
	peer->attr = NULL;
    }

    peer->code = peer->subcode = 0;
    return (peer->attr);
}


/* 
 * Create MRT update message. Take in pointer to
 * buffer of allocated memory, and attr and with and
 * ann prefixes. Return pointer to end of buffer
 * filled in with MRT update message. Set size to size
 * filled in.
 */
u_char *
bgp_create_update_msg (enum MRT_MSG_TYPES type, u_int * size,
		       bgp_attr_t * attr,
		       LINKED_LIST * ll_ann_prefixes,
		       LINKED_LIST * ll_with_prefixes,
		       u_short as, prefix_t * prefix)
{
    u_char *cp, *start_pdu, *end_pdu;
    short len;
    int plen = 4;
    int bgp4plus = DEFAULT_BGP4PLUS_VERSION;

#ifdef HAVE_IPV6
    if (type == MSG_PROTOCOL_BGP4PLUS) {
	assert (attr->gateway == NULL ||
		attr->gateway->prefix->family == AF_INET6);
	plen = 16;
    }
    if (type == MSG_PROTOCOL_BGP4PLUS_01) {
	assert (attr->gateway == NULL ||
		attr->gateway->prefix->family == AF_INET6);
	plen = 16;
	bgp4plus = 1;
    }
#endif /* HAVE_IPV6 */

    cp = start_pdu = NewArray (u_char, MAX_MSG_SIZE);
    end_pdu = start_pdu + MAX_MSG_SIZE;

    /* from */
    if (attr->gateway == NULL)
	cp += (plen + 2);
    else {
	BGP_PUT_SHORT (attr->gateway->AS, cp);
	memcpy (cp, prefix_tochar (attr->gateway->prefix), plen);
	cp += plen;
    }

    /* to */
    if (attr->gateway == NULL)
	cp += (plen + 2);
    else {
	BGP_PUT_SHORT (as, cp);
	if (prefix)
	    memcpy (cp, prefix_tochar (prefix), plen);
	cp += plen;
    }

    len = bgp_create_pdu (ll_with_prefixes, ll_ann_prefixes, attr, 0,
			  NULL, NULL, cp, end_pdu - cp, bgp4plus);

    if (len > 0) {
	*size = (cp + len) - start_pdu;
    }

    return (start_pdu);
}


/* 
 * Given a pointer to allocated memory (cp), fill in memory with BGP
 * attributes in BGP4 packet format, and return pointer to end of 
 * attributes block. If as > 0, add as to start of aspath
 * if cp == NULL, just generates trace outputs
 */
u_char *
bgp_add_attributes (u_char * cp, int cp_len, bgp_attr_t * attr, int as,
		    trace_t * tr)
{
    u_char *start = cp;
    int len = 0;
    int plen;

    /* if pointer is given, check if overflow and return NULL in the case */
    if (cp) {

	if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ORIGIN))
	    len += 4; /* origin */ ;

	if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ASPATH)) {
	    len += 3 /* flag & type & len */ ;
	    plen = aspath_length (attr->aspath, as);
	    if (plen > 255)
		len += 1;
	    len += plen;
	}

	if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_NEXTHOP)) {
	    /* ipv4 next hop */
#ifdef HAVE_IPV6
	    if (attr->nexthop->family == AF_INET ||
	       (attr->nexthop->family == AF_INET6 && attr->nexthop4))
#endif /* HAVE_IPV6 */
	    /* is it a must even in case of ipv6 ? */
	    len += (3 + 4);
	}

	if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_METRIC))
	    len += (3 + 4);
	if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_LOCALPREF))
	    len += (3 + 4);
	if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ATOMICAGG))
	    len += (3 + 0);
	if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_AGGREGATOR))
	    len += (3 + 2 + 4);
	if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_DPA))
	    len += (3 + 2 + 4);
	if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_COMMUNITY)) {
	    assert (attr->community);
	    len += 3;
	    plen = attr->community->len * 4;
	    if (plen > 255)
		len += 1;
	    len += plen;
	}

	if (len > cp_len) {
	    return (NULL);
	}
    }

    trace (TR_PACKET, tr, "attribute:\n");

    /* origin */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ORIGIN)) {
	trace (TR_PACKET, tr, "  ORIGIN: %s\n", origin2string (attr->origin));
	if (cp) {
	    PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_ORIGIN, 1, cp);
	    if (attr->origin != 3) /* Aggregate is 3 internally */
	        BGP_PUT_BYTE (attr->origin, cp);
	    else /* Agrregate route is incomplete */
	        BGP_PUT_BYTE (2, cp);
	}
    }

    /* aspath */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ASPATH)) {
	/* assert (attr->aspath); */
	if (as > 0)
	    trace (TR_PACKET, tr, "  ASPATH: %d %s\n", as,
		   aspath_toa (attr->aspath));
	else
	    trace (TR_PACKET, tr, "  ASPATH: %s\n", aspath_toa (attr->aspath));
	if (cp) {
	    u_char *xp;
	    /* NULL aspath brings a NULL aspath segment with len = 0 */
	    plen = aspath_length (attr->aspath, as);
	    PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_ASPATH, plen, cp);
	    xp = cp;
	    cp = unmunge_aspath (attr->aspath, cp, as);
	    assert (plen == (cp - xp));
	}
    }

    /* next hop */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_NEXTHOP)) {
	assert (attr->nexthop);
#ifdef HAVE_IPV6
	if (attr->nexthop->family == AF_INET6) {
	    if (attr->nexthop4) {
		trace (TR_PACKET, tr,
		       "  NEXT_HOP: %s\n", prefix_toa (attr->nexthop4));
		if (cp) {
		    PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_NEXTHOP, 4, cp);
		    BGP_PUT_NETLONG (prefix_tolong (attr->nexthop4), cp);
		}
	    }
#ifdef notdef
	    else {
		/* I'm not sure but it's safer */
		trace (TR_PACKET, tr,
		       "  NEXT_HOP: 0.0.0.0\n");
		if (cp) {
		    PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_NEXTHOP, 4, cp);
		    BGP_PUT_NETLONG (0, cp);
		}
	    }
#endif
	}
	else
#endif /* HAVE_IPV6 */
	{
	    trace (TR_PACKET, tr,
		   "  NEXT_HOP: %s\n", prefix_toa (attr->nexthop));
	    if (cp) {
		PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_NEXTHOP, 4, cp);
		BGP_PUT_NETLONG (prefix_tolong (attr->nexthop), cp);
	    }
	}
    }

    /* MULTI_EXIT_DISCRIMINATOR: non - mandatory attribute?   */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_METRIC)) {
	trace (TR_PACKET, tr, "  MULTI_EXIT_DISC: %d\n", attr->multiexit);
	if (cp) {
	    PATH_PUT_ATTR (PA_FLAG_OPT, PA4_TYPE_METRIC, 4, cp);
	    BGP_PUT_LONG (attr->multiexit, cp);
	}
    }

    /* LOCAL_PREF: well-known discretionary */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_LOCALPREF)) {
	trace (TR_PACKET, tr, "  LOCAL_PREF: %d\n", attr->local_pref);
	if (cp) {
	    PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_LOCALPREF, 4, cp);
	    BGP_PUT_LONG (attr->local_pref, cp);
	}
    }

    /* Atomic Aggregate */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ATOMICAGG)) {
	trace (TR_PACKET, tr, "  ATOMIC_AGGREGATE\n");
	if (cp) {
	    PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_ATOMICAGG, 0, cp);
	}
    }

    /* Aggregator */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_AGGREGATOR)) {
	assert (attr->aggregator.prefix);
	assert (attr->aggregator.prefix->family == AF_INET);
	trace (TR_PACKET, tr, "  AGGREGATOR: AS%d %s\n",
	       attr->aggregator.as, prefix_toa (attr->aggregator.prefix));
	if (cp) {
	    PATH_PUT_ATTR (PA_FLAG_OPT | PA_FLAG_TRANS, PA4_TYPE_AGGREGATOR, 
			   6, cp);
	    BGP_PUT_SHORT (attr->aggregator.as, cp);
	    BGP_PUT_NETLONG (prefix_tolong (attr->aggregator.prefix), cp);
	}
    }

    /* community */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_COMMUNITY)) {
	assert (attr->community);
	trace (TR_PACKET, tr, "  COMMUNITY: %s\n",
	       community_toa (attr->community));
	if (cp) {
	    PATH_PUT_ATTR (PA_FLAG_OPT | PA_FLAG_TRANS, PA4_TYPE_COMMUNITY,
			   attr->community->len * 4, cp);
	    cp = unmunge_community (attr->community, cp);
	}
    }

    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_DPA)) {
	trace (TR_PACKET, tr, "  DPA: AS%d %d\n",
	       attr->dpa.as, attr->dpa.value);
	if (cp) {
	    PATH_PUT_ATTR (PA_FLAG_OPT | PA_FLAG_TRANS, PA4_TYPE_DPA, 6, cp);
	    BGP_PUT_SHORT (attr->dpa.as, cp);
	    BGP_PUT_LONG (attr->dpa.value, cp);
	}
    }

    if (cp) {
	assert (len == (cp - start));
    }
    return (cp);
}


#ifdef notdef
#define bgp_attr_len(len) (1 /*attr flag*/ + 1 /*attr type*/ + 1 /*len*/ \
			   + (((len) > 255)? 1: 0) /* ext len */ + (len))

static int
bgp_attr_list_len (LINKED_LIST *vp)
{
    u_char *cp;
    int total = 0;
    int flags, code, len;

    assert (vp);
    LL_Iterate (vp, cp) {
        GET_PATH_ATTR (flags, code, len, cp);
	total += bgp_attr_len (len);
    }
    return (total);
}


int
bgp_create_attr_list (LINKED_LIST *vp, bgp_attr_t * attr, int as, trace_t * tr)
{
    int total = 0;
    int len;

    trace (TR_PACKET, tr, "attribute:\n");

    /* origin */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ORIGIN)) {
	trace (TR_PACKET, tr, "  ORIGIN: %s\n", origin2string (attr->origin));
	len = bgp_attr_len (1);
 	total += len;
	if (vp) {
    	    u_char *cp = NewArray (u_char, len);
	    LL_Add (vp, cp);
	    PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_ORIGIN, 1, cp);
	    BGP_PUT_BYTE (attr->origin, cp);
	}
    }

    /* aspath */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ASPATH)) {
	int alen;

	/* assert (attr->aspath); */
	if (as > 0)
	    trace (TR_PACKET, tr, "  ASPATH: %d %s\n", as,
		   aspath_toa (attr->aspath));
	else
	    trace (TR_PACKET, tr, "  ASPATH: %s\n", aspath_toa (attr->aspath));
	alen = aspath_length (attr->aspath, as);
	len = bgp_attr_len (alen);
 	total += len;
	if (vp) {
	    u_char *xp;
	    /* NULL aspath brings a NULL aspath segment with len = 0 */
	    u_char *cp = NewArray (u_char, len);
	    LL_Add (vp, cp);
	    xp = cp;
	    PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_ASPATH, alen, cp);
	    cp = unmunge_aspath (attr->aspath, cp, as);
	    assert (len == (cp - xp));
	}
    }

    /* next hop */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_NEXTHOP)) {
	assert (attr->nexthop);
	len = bgp_attr_len (4);
 	total += len;
#ifdef HAVE_IPV6
	if (attr->nexthop->family == AF_INET6) {
	    if (attr->nexthop4) {
		trace (TR_PACKET, tr,
		       "  NEXT_HOP: %s\n", prefix_toa (attr->nexthop4));
		if (vp) {
	    	    u_char *cp = NewArray (u_char, len);
	    	    LL_Add (vp, cp);
		    PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_NEXTHOP, 4, cp);
		    BGP_PUT_NETLONG (prefix_tolong (attr->nexthop4), cp);
		}
	    }
#ifdef notdef
	    else {
		/* I'm not sure but it's safer */
		trace (TR_PACKET, tr,
		       "  NEXT_HOP: 0.0.0.0\n");
		if (vp) {
		    u_char *cp = NewArray (u_char, len);
	    	    LL_Add (vp, cp);
		    PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_NEXTHOP, 4, cp);
		    BGP_PUT_NETLONG (0, cp);
		}
	    }
#endif
	}
	else
#endif /* HAVE_IPV6 */
	{
	    trace (TR_PACKET, tr,
		   "  NEXT_HOP: %s\n", prefix_toa (attr->nexthop));
	    if (vp) {
	    	u_char *cp = NewArray (u_char, len);
	    	LL_Add (vp, cp);
		PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_NEXTHOP, 4, cp);
		BGP_PUT_NETLONG (prefix_tolong (attr->nexthop), cp);
	    }
	}
    }


    /* MULTI_EXIT_DISCRIMINATOR: non - mandatory attribute?   */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_METRIC)) {
	trace (TR_PACKET, tr, "  MULTI_EXIT_DISC: %d\n", attr->multiexit);
	len = bgp_attr_len (4);
 	total += len;
	if (vp) {
	    u_char *cp = NewArray (u_char, len);
	    LL_Add (vp, cp);
	    PATH_PUT_ATTR (PA_FLAG_OPT, PA4_TYPE_METRIC, 4, cp);
	    BGP_PUT_LONG (attr->multiexit, cp);
	}
    }

    /* LOCAL_PREF: well-known discretionary */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_LOCALPREF)) {
	trace (TR_PACKET, tr, "  LOCAL_PREF: %d\n", attr->local_pref);
	len = bgp_attr_len (4);
 	total += len;
	if (vp) {
	    u_char *cp = NewArray (u_char, len);
	    LL_Add (vp, cp);
	    PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_LOCALPREF, 4, cp);
	    BGP_PUT_LONG (attr->local_pref, cp);
	}
    }

    /* Atomic Aggregate */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ATOMICAGG)) {
	trace (TR_PACKET, tr, "  ATOMIC_AGGREGATE\n");
	len = bgp_attr_len (0);
 	total += len;
	if (vp) {
	    u_char *cp = NewArray (u_char, len);
	    LL_Add (vp, cp);
	    PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_ATOMICAGG, 0, cp);
	}
    }

    /* Aggregator */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_AGGREGATOR)) {
	assert (attr->aggregator.prefix);
	assert (attr->aggregator.prefix->family == AF_INET);
	trace (TR_PACKET, tr, "  AGGREGATOR: AS%d %s\n",
	       attr->aggregator.as, prefix_toa (attr->aggregator.prefix));
	len = bgp_attr_len (6);
 	total += len;
	if (vp) {
	    u_char *cp = NewArray (u_char, len);
	    LL_Add (vp, cp);
	    PATH_PUT_ATTR (PA_FLAG_OPT | PA_FLAG_TRANS, PA4_TYPE_AGGREGATOR, 
			   6, cp);
	    BGP_PUT_SHORT (attr->aggregator.as, cp);
	    BGP_PUT_NETLONG (prefix_tolong (attr->aggregator.prefix), cp);
	}
    }

    /* community */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_COMMUNITY)) {
	int clen;
	assert (attr->community);
	clen = attr->community->len * 4;
	len = bgp_attr_len (clen);
 	total += len;
	trace (TR_PACKET, tr, "  COMMUNITY: %s\n",
	       community_toa (attr->community));
	if (vp) {
	    u_char *xp;
	    u_char *cp = NewArray (u_char, len);
	    LL_Add (vp, cp);
	    xp = cp;
	    PATH_PUT_ATTR (PA_FLAG_OPT | PA_FLAG_TRANS, PA4_TYPE_COMMUNITY,
			   clen, cp);
	    cp = unmunge_community (attr->community, cp);
	    assert (cp - xp == len);
	}
    }

    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_DPA)) {
	trace (TR_PACKET, tr, "  DPA: AS%d %d\n",
	       attr->dpa.as, attr->dpa.value);
	len = bgp_attr_len (6);
 	total += len;
	if (vp) {
	    u_char *cp = NewArray (u_char, len);
	    LL_Add (vp, cp);
	    PATH_PUT_ATTR (PA_FLAG_OPT | PA_FLAG_TRANS, PA4_TYPE_DPA, 6, cp);
	    BGP_PUT_SHORT (attr->dpa.as, cp);
	    BGP_PUT_LONG (attr->dpa.value, cp);
	}
    }

    if (vp) {
	assert (total == bgp_attr_list_len (vp));
    }
    return (total);
}
#endif


void
bgp_print_attr (bgp_attr_t * attr)
{
    char tmpx[MAXLINE];

    if (attr->gateway && (attr->gateway->AS > 0))
	printf ("FROM: %s\n", gateway_toa (tmpx, attr->gateway));

    /* may be a withdraw attr with just peer info */
    if (attr->aspath == NULL)
	return;

    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ORIGIN))
	printf ("ORIGIN: %-12s\n", s_origins[attr->origin]);
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ASPATH))
	printf ("ASPATH: %s\n", aspath_toa (attr->aspath));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_NEXTHOP))
	printf ("NEXT_HOP: %s\n", prefix_toa (attr->nexthop));
#ifdef HAVE_IPV6
    if (attr->link_local)
	printf ("NEXT_HOP: %s\n", prefix_toa (attr->link_local));
    if (attr->nexthop4)
	printf ("NEXT_HOP: %s\n", prefix_toa (attr->nexthop4));
#endif /* HAVE_IPV6 */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_METRIC))
	printf ("MULTI_EXIT_DISC: %ld\n", attr->multiexit);
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_LOCALPREF))
	printf ("LOCAL_PREF: %ld\n", attr->local_pref);
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ATOMICAGG))
	printf ("ATOMIC_AGGREGATE\n");
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_AGGREGATOR))
	printf ("AGGREGATOR: AS%d %s\n", attr->aggregator.as,
		prefix_toa (attr->aggregator.prefix));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_COMMUNITY))
	printf ("COMMUNITY: %s\n", community_toa (attr->community));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_DPA))
	printf ("DPA: AS%d %ld\n", attr->dpa.as, attr->dpa.value);
}


void
bgp_trace_attr2 (bgp_attr_t * attr, int as, trace_t * tr)
{
    bgp_add_attributes (NULL, 0, attr, as, tr);
#ifdef HAVE_IPV6
    if (attr->link_local)
	trace (TR_PACKET, tr, "  NEXT_HOP: %s\n", 
	       prefix_toa (attr->link_local));
    if (attr->nexthop && attr->nexthop->family == AF_INET6)
	trace (TR_PACKET, tr, "  NEXT_HOP: %s\n", 
	       prefix_toa (attr->nexthop));
#endif /* HAVE_IPV6 */
}


char *
bgp_attr_toa (bgp_attr_t * attr)
{
    char *tmpx;
    char *cp;

    THREAD_SPECIFIC_STRAGE (tmpx);

    sprintf (tmpx, "%s|%s|%s|%d|%d|%s|%s|",
	     aspath_toa (attr->aspath),
	     s_origins[attr->origin],
	     attr->nexthop ? prefix_toa (attr->nexthop) : "none",
	     (int) attr->local_pref,
	     (int) attr->multiexit,
	     community_toa (attr->community),
	     BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ATOMICAGG) ? "AG" : "NAG");

    /* I don't think we can call prefix_toa twice in the same call */
    cp = tmpx + strlen (tmpx);
    sprintf (cp, "%s|",  
	     attr->aggregator.prefix ? prefix_toa (attr->aggregator.prefix) : "");

    return (tmpx);
}


bgp_attr_t *
bgp_new_attr (int type)
{
    bgp_attr_t *attr;

    attr = New (bgp_attr_t);
    attr->type = type;
    attr->ref_count = 1;
    attr->attribs = 0;
    attr->nexthop = NULL;
    attr->gateway = NULL;
    attr->nodelete = 0;
#ifdef notdef
    BGP4_BIT_SET (attr->attribs, PA4_TYPE_ORIGIN);
    if (type == PROTO_BGP)
	attr->origin = 1;
    else if (type == PROTO_RIPNG || type == PROTO_RIP)
	attr->origin = 0;
    else if (type == PROTO_KERNEL || type == PROTO_CONNECTED ||
	     type == PROTO_STATIC)
	attr->origin = 2;

    /* XXX we only can specify the origin along with redistribute */
    if (BIT_TEST (MRT->redist[type], (1 << PROTO_BGP)))
	attr->origin = MRT->origin[type];

#endif
    num_active_bgp_attr++;
    pthread_mutex_init (&attr->mutex_lock, NULL);

    return (attr);
}


bgp_attr_t *
bgp_ref_attr (bgp_attr_t * attr)
{
    pthread_mutex_lock (&attr->mutex_lock);
    assert (attr->ref_count > 0);
    attr->ref_count++;
    pthread_mutex_unlock (&attr->mutex_lock);
    return (attr);
}


void
bgp_deref_attr (bgp_attr_t * attr)
{
    pthread_mutex_lock (&attr->mutex_lock);

    assert (attr->ref_count > 0);
    if (--attr->ref_count <= 0) {

	if (attr->nodelete == 0) {
	    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_NEXTHOP)) {
		assert (attr->nexthop);
		/* nexthop can be nexthop in attribute or one from 4+ extension */
		Deref_Prefix (attr->nexthop);
	    }
	    else {
		assert (attr->nexthop == NULL);
	    }
	    if (attr->nexthop2)
		Deref_Prefix (attr->nexthop2);
	    /* if (attr->gateway) Delete (attr->gateway); */
	    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ASPATH)) {
	        if (attr->aspath)
		    Delete_ASPATH (attr->aspath);
	    }
	    else {
		assert (attr->aspath == NULL);
	    }
	    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_AGGREGATOR)) {
		assert (attr->aggregator.prefix);
		Deref_Prefix (attr->aggregator.prefix);
	    }
	    else {
		assert (attr->aggregator.prefix == NULL);
	    }
	    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_COMMUNITY)) {
		assert (attr->community);
		Delete_community (attr->community);
	    }
	    else {
		assert (attr->community == NULL);
	    }
#ifdef HAVE_IPV6
	    if (attr->nexthop4) {
		Deref_Prefix (attr->nexthop4);
	    }
	    if (attr->link_local) {
		Deref_Prefix (attr->link_local);
	    }
#endif /* HAVE_IPV6 */
	}
	pthread_mutex_destroy (&attr->mutex_lock);
        num_active_bgp_attr--;
	Delete (attr);
	return;
    }

    pthread_mutex_unlock (&attr->mutex_lock);
}


/* 
 * return 1 if BGP attributes are the same 
 */
int
bgp_compare_attr (bgp_attr_t * a1, bgp_attr_t * a2)
{
    if (a1->gateway->AS != a2->gateway->AS)
	return (-1);

    if (a1->attribs ^ a2->attribs)
	return (-1);

    if (BGP4_BIT_TEST (a1->attribs, PA4_TYPE_ORIGIN) &&
        a1->origin != a2->origin)
	return (-1);

    if (BGP4_BIT_TEST (a1->attribs, PA4_TYPE_ASPATH) &&
        compare_aspaths (a1->aspath, a2->aspath) < 0)
	return (-1);

    if (BGP4_BIT_TEST (a1->attribs, PA4_TYPE_NEXTHOP) &&
	prefix_compare2 (a1->nexthop, a2->nexthop) != 0)
	return (-1);

#ifdef HAVE_IPV6
    if (a1->nexthop4 || a2->nexthop4) {
	if (!a1->nexthop4 || !a2->nexthop4)
	    return (-1);
	if (prefix_compare2 (a1->nexthop4, a2->nexthop4) != 0)
	    return (-1);
    }
    if (a1->link_local || a2->link_local) {
	if (!a1->link_local || !a2->link_local)
	    return (-1);
	if (prefix_compare2 (a1->link_local, a2->link_local) != 0)
	    return (-1);
    }
#endif /* HAVE_IPV6 */
    if (BGP4_BIT_TEST (a1->attribs, PA4_TYPE_METRIC) &&
	a1->multiexit != a2->multiexit)
	return (-1);
    if (BGP4_BIT_TEST (a1->attribs, PA4_TYPE_LOCALPREF) &&
	a1->local_pref != a2->local_pref)
	return (-1);
    if (BGP4_BIT_TEST (a1->attribs, PA4_TYPE_AGGREGATOR) &&
	(a1->aggregator.as != a2->aggregator.as ||
	prefix_compare2 (a1->aggregator.prefix, a2->aggregator.prefix) != 0))
	return (-1);
    if (BGP4_BIT_TEST (a1->attribs, PA4_TYPE_COMMUNITY) &&
	community_compare (a1->community, a2->community) < 0)
	return (-1);
    if (BGP4_BIT_TEST (a1->attribs, PA4_TYPE_DPA) &&
	(a1->dpa.as != a2->dpa.as || a1->dpa.value != a2->dpa.value))
	return (-1);

    return (1);
}


/* 
 * here is where we actually build update packets and 
 * schedule them to be sent off. 
 * NOTE: It is is the responsibility of bgp peer to delete buffer memory!
 */

/* if as > 0, prepend as at the beginning of as path
   if as == 0, nothing done */
int
bgp_create_pdu (LINKED_LIST * ll_with_prefixes, LINKED_LIST * ll_ann_prefixes,
		bgp_attr_t * attr, int as,
		/* these argument are used by bgp daemon */
		bgp_peer_t * peer, int (*fn) (),
		/* these argument are used by tools like route_atob */
		u_char * pdu_memory, int pdu_len, int bgp4plus)
{
    prefix_t *p_with_prefix = NULL, *p_ann_prefix = NULL;
#ifdef HAVE_IPV6
    prefix_t *p_with_prefix6 = NULL, *p_ann_prefix6 = NULL;
#endif /* HAVE_IPV6 */
    trace_t *tr = MRT->trace;

    if (peer) {
	tr = peer->trace;
        bgp4plus = (BIT_TEST (peer->options, BGP_BGP4PLUS_01)? 1: 0);
    }
    else {
	assert (pdu_memory);
    }

    if (ll_with_prefixes)
	p_with_prefix = LL_GetHead (ll_with_prefixes);
    if (ll_ann_prefixes)
	p_ann_prefix = LL_GetHead (ll_ann_prefixes);

    while (p_with_prefix || p_ann_prefix 
#ifdef HAVE_IPV6
	    || p_with_prefix6 || p_ann_prefix6
#endif /* HAVE_IPV6 */
	    ) {

        u_char *p_total_attrib_len, *p_unfeasible_len;
	u_char *cp, *start_pdu, *maxend;
	int overflow = 0;
        int announce4 = 0;
        int num_prefixes = 0;
	int first;
        u_char *start_attr;
#ifdef HAVE_IPV6
        u_char *start_v6;
#endif /* HAVE_IPV6 */

        trace (TR_DEBUG, tr, "send ---- start\n");

	if (pdu_memory && pdu_len > 0) {
	    cp = start_pdu = pdu_memory;
	    maxend = cp + pdu_len;
	}
	else {
	    cp = start_pdu = NewArray (u_char, BGPMAXPACKETSIZE);
	    maxend = cp + BGPMAXPACKETSIZE - BGP_HEADER_LEN;
	}

	p_unfeasible_len = cp;
	/* skip unfeasible route length for now */
	BGP_PUT_SHORT (0, cp);

	/* 
	 * search IPv4 withdrawn prefixes and add them first
	 */
	first = 1;
	while (p_with_prefix) {

	    if (p_with_prefix->family != AF_INET) {
#ifdef HAVE_IPV6
	        if (p_with_prefix->family == AF_INET6) {
		    if (p_with_prefix6 == NULL)
		        p_with_prefix6 = p_with_prefix;
		}
#endif /* HAVE_IPV6 */
		p_with_prefix = LL_GetNext (ll_with_prefixes, p_with_prefix);
		continue;
	    }

	    if (cp + BGP_PREFIX_LEN (p_with_prefix->bitlen) + 1 <= maxend) {
		if (first) {
	    	    trace (TR_PACKET, tr, "send withdraw:\n");
		    first = 0;
		}
		BGP_PUT_PREFIX (p_with_prefix->bitlen,
				prefix_tochar (p_with_prefix), cp);
		trace (TR_PACKET, tr, "  %s\n", prefix_toax (p_with_prefix));
		p_with_prefix = LL_GetNext (ll_with_prefixes, p_with_prefix);

		num_prefixes++;
		if (peer->maximum_prefix > 0 && 
			num_prefixes >= peer->maximum_prefix) {
		    overflow++;
		    break;
		}
	    }
	    else {
		overflow++;
		break;
	    }
	}
	assert (cp <= maxend);

	if (cp - start_pdu - 2 > 0) {
	    trace (TR_PACKET, tr, "send withdraw len = %d\n",
	           cp - start_pdu - 2);
	    /* put in withdrawn routes length */
	    BGP_PUT_SHORT (cp - start_pdu - 2, p_unfeasible_len);
	    p_unfeasible_len -= 2;
	}

	/* skip total path attribute length for now */
	p_total_attrib_len = cp;
	BGP_PUT_SHORT (0, cp);
	start_attr = cp;

	/* if there is at least one v4 or v6 announce route,
	   emit the attribute */
	if (!overflow && p_ann_prefix) {

	    cp = bgp_add_attributes (cp, maxend - cp, attr, as, tr);

	    if (cp == NULL) {
		trace (TR_ERROR, tr, "too large attributes\n");
		if (!(pdu_memory && pdu_len > 0))
		    Delete (start_pdu);
		return (-1);
	    }

	    trace (TR_PACKET, tr, "send attribute len updated = %d\n",
		   cp - start_attr);

	    /* Now put in attribute length */
	    BGP_PUT_SHORT (cp - start_attr, p_total_attrib_len);
	    p_total_attrib_len -= 2;
        }

	/* 
	 * add IPv4 announce if no more ipv4 withdraw
	 */
	while (!overflow && p_ann_prefix) {

	    if (p_ann_prefix->family != AF_INET) {
#ifdef HAVE_IPV6
	        if (p_ann_prefix->family == AF_INET6) {
		    if (p_ann_prefix6 == NULL)
		        p_ann_prefix6 = p_ann_prefix;
		}
#endif /* HAVE_IPV6 */
		p_ann_prefix = LL_GetNext (ll_ann_prefixes, p_ann_prefix);
		continue;
	    }

	    if (cp + BGP_PREFIX_LEN (p_ann_prefix->bitlen) + 1 <= maxend) {
		if (announce4 == 0)
	    	    trace (TR_PACKET, tr, "send announce:\n");

		BGP_PUT_PREFIX (p_ann_prefix->bitlen,
				prefix_tochar (p_ann_prefix), cp);
		trace (TR_PACKET, tr, "  %s\n", prefix_toax (p_ann_prefix));
		announce4++;
		p_ann_prefix = LL_GetNext (ll_ann_prefixes, p_ann_prefix);

		num_prefixes++;
		if (peer->maximum_prefix > 0 && 
			num_prefixes >= peer->maximum_prefix) {
		    overflow++;
		    break;
		}
	    }
	    else {
		overflow++;
		break;
	    }
	}
	assert (cp <= maxend);

#ifdef HAVE_IPV6
	/* 
	 * add IPv6 announce routes if no more ipv4 withdrawn & announce routes 
	 */
	if (!overflow && announce4 == 0 /* can't mix now */
		      && p_ann_prefix6) {

	    prefix_t *pref = p_ann_prefix6;
	    int xlen, plen = 0;
	    int family = BGP_DEFAULT_IPV6_FAMILY;
	    int subfamily = BGP_DEFAULT_IPV6_SUBFAMILY;
    	    int nhalen = 16;

	    if (attr->link_local && attr->link_local->family == AF_INET6)
		nhalen += 16;

	    while (pref) {

	        if (pref->family != AF_INET6) {
		    pref = LL_GetNext (ll_ann_prefixes, pref);
		    continue;
		}

		plen += 1;	/* prefix's length field */
		plen += BGP_PREFIX_LEN (pref->bitlen);

		if (bgp4plus == 0)
		    xlen = plen + 4 + nhalen /* next hop */  + 1 + 2;
		else
		    xlen = plen + 4 + nhalen /* next hop */  + 1;
		if (cp + 3 + xlen + (xlen > 255 ? 1 : 0) > maxend) {
		    plen -= BGP_PREFIX_LEN (pref->bitlen);
		    plen -= 1;
		    overflow++;
		    break;
		}

		pref = LL_GetNext (ll_ann_prefixes, pref);

		num_prefixes++;
		if (peer->maximum_prefix > 0 && 
			num_prefixes >= peer->maximum_prefix) {
		    overflow++;
		    break;
		}
	    }

if (plen > 0) {

	    if (bgp4plus == 0)
		xlen = plen + 4 + nhalen + 1 + 2;
	    else
		xlen = plen + 4 + nhalen + 1;

	    /* estimation is required because if xlen > 255,
	       length field becomes 2 octets */
	    PATH_PUT_ATTR (PA_FLAG_OPT, PA4_TYPE_MPREACHNLRI, xlen, cp);
	    start_v6 = cp; /* to check later */

	    BGP_PUT_SHORT (family, cp);	/* family */
	    BGP_PUT_BYTE (subfamily, cp);	/* subfamily */
	    BGP_PUT_BYTE (nhalen, cp);

	    trace (TR_PACKET, tr,
		   "send announce family %d subfamily %d nhalen %d\n",
		   family, subfamily, nhalen);

	    trace (TR_PACKET, tr,
		   "send announce draft version %d\n",
		   bgp4plus);

	    if (attr->nexthop && attr->nexthop->family == AF_INET6) {
		trace (TR_PACKET, tr,
		       "  NEXT_HOP: %s\n", prefix_toa (attr->nexthop));
		memcpy (cp, prefix_tochar (attr->nexthop), 16);
	    }
	    else {
		trace (TR_PACKET, tr, "  NEXT_HOP: ::\n");
		memset (cp, 0, 16);
	    }
	    cp += 16;

	    if (nhalen == 32) {
		trace (TR_PACKET, tr,
		       "  NEXT_HOP: %s\n", prefix_toa (attr->link_local));
		memcpy (cp, prefix_tochar (attr->link_local), 16);
	        cp += 16;
	    }

#ifdef notdef
	    trace (TR_PACKET, tr,
		   "send announce number of SNAPs = %d\n", 0);
#endif
	    BGP_PUT_BYTE (0, cp);	/* Number of SNPAs */

	    if (bgp4plus == 0) {
		trace (TR_PACKET, tr,
		       "send announce draft 0 len %d\n", plen);
		BGP_PUT_SHORT (plen, cp);
	    }

	    trace (TR_PACKET, tr, "send announce:\n");

	    while (p_ann_prefix6 != pref) {
	        if (p_ann_prefix6->family == AF_INET6) {
		    BGP_PUT_PREFIX6 (p_ann_prefix6->bitlen,
				     prefix_tochar (p_ann_prefix6), cp);
		    trace (TR_PACKET, tr, "  %s\n", 
			   prefix_toax (p_ann_prefix6));
		}
		p_ann_prefix6 = LL_GetNext (ll_ann_prefixes, p_ann_prefix6);
	    }
	    assert (cp <= maxend);
	    assert (start_v6 + xlen == cp);
	    BGP_PUT_SHORT (cp - start_attr, p_total_attrib_len);
	    p_total_attrib_len -= 2;
	    trace (TR_PACKET, tr, "send attribute len updated = %d\n",
		   cp - start_attr);
}
	}


	/*
	 * add IPv6 withdrawn prefixes as a multi-protocol attribute
	 * this should follow ipv6 announce to avoid sorting by attr code
	 */
	if (!overflow && announce4 == 0 && p_with_prefix6) {
	    prefix_t *pref = p_with_prefix6;
	    int xlen, plen = 0;
	    int family = BGP_DEFAULT_IPV6_FAMILY;
	    int subfamily = BGP_DEFAULT_IPV6_SUBFAMILY;

#undef CISCO
#ifdef CISCO
	    if (emit_attr == 0) {
	        PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_ORIGIN, 1, cp);
	        BGP_PUT_BYTE (0, cp);

	        PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_ASPATH, 2, cp);
                BGP_PUT_BYTE (PA_PATH_SEQ, cp);
                BGP_PUT_BYTE (0, cp);

	        PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_NEXTHOP, 4, cp);
	        BGP_PUT_NETLONG (0, cp);

                PATH_PUT_ATTR (PA_FLAG_TRANS, PA4_TYPE_LOCALPREF, 4, cp);
                BGP_PUT_LONG (0, cp);
	    }
#endif /* CISCO */

	    /* estimate the length needed */
	    while (pref) {

	        if (pref->family != AF_INET6) {
		    pref = LL_GetNext (ll_with_prefixes, pref);
		    continue;
		}

		plen += 1;
		plen += BGP_PREFIX_LEN (pref->bitlen);

		if (bgp4plus == 0)
		    xlen = plen + 3 + 2;
		else
		    xlen = plen + 3;
		if (cp + 3 + xlen + (xlen > 255 ? 1 : 0) > maxend) {
		    plen -= BGP_PREFIX_LEN (pref->bitlen);
		    plen -= 1;
		    overflow++;
		    break;
		}
		pref = LL_GetNext (ll_with_prefixes, pref);

		num_prefixes++;
		if (peer->maximum_prefix > 0 && 
			num_prefixes >= peer->maximum_prefix) {
		    overflow++;
		    break;
		}
	    }

	    if (bgp4plus == 0)
		xlen = plen + 3 + 2;
	    else
		xlen = plen + 3;
	    PATH_PUT_ATTR (PA_FLAG_OPT, PA4_TYPE_MPUNRNLRI, xlen, cp);
	    start_v6 = cp;

	    BGP_PUT_SHORT (family, cp);		/* family */
	    BGP_PUT_BYTE (subfamily, cp);	/* subfamily */
	    trace (TR_PACKET, tr,
		   "send withdraw family %d subfamily %d\n",
		   family, subfamily);

	    trace (TR_PACKET, tr, "send withdraw draft version %d\n",
		   bgp4plus);
	    if (bgp4plus == 0) {
		trace (TR_PACKET, tr,
		       "send withdraw draft 0 len %d\n", plen);
		BGP_PUT_SHORT (plen, cp);
	    }

	    trace (TR_PACKET, tr, "send withdraw:\n");
	    while (p_with_prefix6 != pref) {
	        if (p_with_prefix6->family == AF_INET6) {
		    BGP_PUT_PREFIX6 (p_with_prefix6->bitlen,
				     prefix_tochar (p_with_prefix6), cp);
		    trace (TR_PACKET, tr, "  %s\n",
		           prefix_toax (p_with_prefix6));
		}
		p_with_prefix6 = LL_GetNext (ll_with_prefixes, p_with_prefix6);
	    }
	    assert (cp <= maxend);
	    assert (start_v6 + xlen == cp);
	    BGP_PUT_SHORT (cp - start_attr, p_total_attrib_len);
	    p_total_attrib_len -= 2;
	    trace (TR_PACKET, tr, "send attribute len updated = %d\n",
		   cp - start_attr);
	}
#endif /* HAVE_IPV6 */

	trace (TR_PACKET, tr,
	       "send ---- end (%d prefixes, len = %d)\n",
	       num_prefixes, cp - start_pdu);
	if (pdu_memory && pdu_len > 0) {
	    if (overflow)
		return (-1);
	    return (cp - start_pdu);
	}
	else if (peer && fn) {
	    /* schedule for sending it */
	    (*fn) (peer, cp - start_pdu, start_pdu);
	}
    }
    return (0);
}


bgp_attr_t *
bgp_copy_attr (bgp_attr_t * attr)
{
    bgp_attr_t *new_attr;

    assert (attr);
    new_attr = bgp_new_attr (attr->type);

    new_attr->attribs = attr->attribs;
    new_attr->gateway = attr->gateway;
    new_attr->tag = attr->tag;

    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_NEXTHOP))
	new_attr->nexthop = Ref_Prefix (attr->nexthop);
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ASPATH))
	new_attr->aspath = aspath_copy (attr->aspath);
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ORIGIN))
	new_attr->origin = attr->origin;

    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_METRIC))
	new_attr->multiexit = attr->multiexit;
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_LOCALPREF))
	new_attr->local_pref = attr->local_pref;
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_DPA))
	new_attr->dpa = attr->dpa;
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_COMMUNITY)) {
	assert (attr->community);
	new_attr->community = community_copy (attr->community);
    }
#ifdef HAVE_IPV6
    if (attr->nexthop4)
	new_attr->nexthop4 = Ref_Prefix (attr->nexthop4);
    if (attr->link_local)
	new_attr->link_local = Ref_Prefix (attr->link_local);
#endif /* HAVE_IPV6 */
    if (attr->nexthop2)
	new_attr->nexthop2 = Ref_Prefix (attr->nexthop2);

    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_AGGREGATOR)) {
	new_attr->aggregator.as = attr->aggregator.as;
	new_attr->aggregator.prefix = Ref_Prefix (attr->aggregator.prefix);
    }
    new_attr->home_AS = attr->home_AS;

    return (new_attr);
}
