/* 
 * $Id: config.c,v 1.28 1998/07/17 18:47:48 masaki Exp $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <stdarg.h>

#include <mrt.h>
#include <trace.h>
#include <io.h>
#include <select.h>
#include <interface.h>
#include <rip.h>
#ifdef HAVE_IPV6
#include <ripng.h>
#endif /* HAVE_IPV6 */
#include <bgp.h>
#include <config_file.h>

#include "config.h"
#include "rib.h"

config_mrtd_t *CONFIG_MRTD;
int BGPSIM = 0;
int running = 0;

/* Comments...
 * Config can work from a file or from user configuration via a
 * socket.
 */


static void
get_config_ip_route (static_route_t * static_route)
{
    char options[MAXLINE], *cp = options;

    strcpy (cp, "");
    if (static_route->pref >= 0) {
	sprintf (cp, " %d", static_route->pref);
	cp += strlen (cp);
    }
    if (static_route->interface) {
	sprintf (cp, " %s", static_route->interface->name);
	cp += strlen (cp);
    }
    config_add_output ("route %s %s%s\n",
		       prefix_toax (static_route->prefix),
		       prefix_toa (static_route->nexthop),
		       options);
}


/* 
 * config a static ip route
 */
static int
config_ip_route (uii_connection_t * uii, prefix_t * prefix,
		 prefix_t * nexthop, char *options)
{
    static_route_t *static_route;
    interface_t *interface = NULL;
    char name[MAXLINE];
    int pref = -1;

    LL_Iterate (CONFIG_MRTD->ll_static_routes, static_route) {
	if (prefix_compare (prefix, static_route->prefix)) {
    	    config_notice (TR_PARSE, uii,
		   	    "static route %s already defined\n",
		   	    prefix_toax (prefix));
	    Deref_Prefix (prefix);
	    Deref_Prefix (nexthop);
	    if (options) Delete (options);
	    return (-1);
	}
    }

    strcpy (name, "");
    if (options) {
	if (parse_line (options, "%n %d", name, &pref) <= 0) {
	    if (parse_line (options, "%d %n", &pref, name) <= 0) {
		Delete (options);
	    	return (-1);
	    }
	}
	if (name[0]) {
	    interface = find_interface_byname (name);
	    if (interface == NULL) {
    	        config_notice (TR_ERROR, uii, "interface %s not found\n", name);
		return (-1);
	    }
	}
    }

    config_notice (TR_PARSE, uii,
		   "static route %s via %s pref %d on %s\n",
		   prefix_toax (prefix), prefix_toa (nexthop), pref, name);

    if (running)
	add_route_to_rib (PROTO_STATIC, prefix, nexthop, interface, 
			 (pref >= 0)? pref: STATIC_PREF, 0);
    /* XXX error handling ? */

    static_route = New (static_route_t);
    static_route->prefix = prefix;
    static_route->nexthop = nexthop;
    static_route->interface = interface;
    static_route->pref = pref;
    LL_Add (CONFIG_MRTD->ll_static_routes, static_route);
    config_add_module (0, "route", get_config_ip_route, static_route);
    if (options) Delete (options);

    return (1);
}


static int
no_config_ip_route (uii_connection_t * uii, prefix_t * prefix)
{
    static_route_t *static_route;

    if (CONFIG_MRTD->ll_static_routes == NULL) {
	config_notice (TR_PARSE, uii, "No matching route to delete for %s\n",
		       prefix_toax (prefix));
	Deref_Prefix (prefix);
	return (0);
    }

    LL_Iterate (CONFIG_MRTD->ll_static_routes, static_route) {
	if (prefix_compare (static_route->prefix, prefix)) {

	    if (running)
		del_route_from_rib (PROTO_STATIC, prefix);
	    /* XXX error handling ? */

	    config_del_module (0, "route", get_config_ip_route,
			       static_route);
	    config_notice (TR_PARSE, uii,
			   "CONFIG static route %s via %s being deleted\n",
			   prefix_toax (prefix),
			   prefix_toa (static_route->nexthop));
	    Deref_Prefix (static_route->prefix);
	    Deref_Prefix (static_route->nexthop);
	    Delete (static_route);
	    LL_Remove (CONFIG_MRTD->ll_static_routes, static_route);
	    Deref_Prefix (prefix);
	    return (1);
	}
    }
    config_notice (TR_PARSE, uii, "No matching route to delete");
    return (0);
}


typedef struct _dlist_t {
    int num;
    char *in_or_out;
    interface_t *interface;
} dlist_t;


static void
get_config_distribute_list (dlist_t * dlist)
{
    config_add_output ("distribute-list %d %s %s\n",
		       dlist->num, dlist->in_or_out,
		       dlist->interface ? dlist->interface->name : "");
}


static int
config_distribute_list (uii_connection_t * uii,
			int num, char *in_or_out, int opt_num, char *name)
{
    interface_t *interface;
    dlist_t *dlist;

    if (num < 0 && num >= MAX_ALIST) {
	config_notice (TR_ERROR, uii,
		       "invalid access-list number (%d)\n", num);
	Delete (in_or_out);
	if (opt_num > 0)
	    Delete (name);
	return (-1);
    }

    if (opt_num <= 0)
	name = "";
    config_notice (TR_PARSE, uii,
		 "CONFIG distribute-list %d %s %s\n", num, in_or_out, name);

    if (*name) {
	if ((interface = find_interface_byname (name)) == NULL) {
	    config_notice (TR_ERROR, uii,
			   "unkown interface %s\n", name);
	    Delete (in_or_out);
	    Delete (name);
	    return (-1);
	}
	if (strcasecmp (in_or_out, "in") == 0)
	    interface->dlist_in[CONFIG_MRTD->protocol] = num;
	else
	    interface->dlist_out[CONFIG_MRTD->protocol] = num;

	dlist = New (dlist_t);
	dlist->num = num;
	dlist->in_or_out = in_or_out;
	dlist->interface = interface;
	config_add_module (0, "distribute-list", get_config_distribute_list,
			   dlist);
	Delete (name);
    }
    else {
	LL_Iterate (INTERFACE_MASTER->ll_interfaces, interface) {
	    if (strcasecmp (in_or_out, "in") == 0)
		interface->dlist_in[CONFIG_MRTD->protocol] = num;
	    else
		interface->dlist_out[CONFIG_MRTD->protocol] = num;
	}

	dlist = New (dlist_t);
	dlist->num = num;
	dlist->in_or_out = in_or_out;
	dlist->interface = NULL;
	config_add_module (0, "distribute-list", get_config_distribute_list,
			   dlist);
    }
    return (1);
}


static void
get_config_aspath_filter (int num)
{
    char *cp = as_access_list_toa (num);
    if (cp) {
	config_add_output ("%s", cp);
	Delete (cp);
    }
}


static int
config_ip_as_filter (uii_connection_t * uii, int num, char *permit_or_deny,
		     char *re)
{
    int permit = 0;

    if (strcasecmp ("permit", permit_or_deny) == 0)
	permit = 1;

    if (num < 0 && num >= MAX_AS_ALIST) {
	config_notice (TR_ERROR, uii,
		   "invalid as-path access-list number (%d)\n", num);
	Delete (permit_or_deny);
	Delete (re);
	return (-1);
    }

    if (add_as_access_list (num, re, permit) > 0) {
	config_notice (TR_TRACE, uii,
		       "CONFIG as-path access-list %d %s %s\n",
		       num, (permit) ? "permit" : "deny", re);
	config_add_module (0, "as-path access-list", get_config_aspath_filter,
			   (void *) num);
    }
    else {
        Delete (permit_or_deny);
        Delete (re);
        return (-1);
    }
    Delete (permit_or_deny);
    Delete (re);
    return (1);
}


/* config_interface
 * start interface configuration 
 */
static void
get_config_interface (interface_t * interface)
{
    config_add_output ("interface %s\n", interface->name);
}


static int
config_interface (uii_connection_t * uii, char *name)
{
    interface_t *interface, *ip;

    if ((interface = find_interface_byname (name)) == NULL) {
	config_notice (TR_ERROR, uii,
		       "INTERFACE %s not found\n", name);
	Delete (name);
	return (-1);
    }
    config_notice (TR_TRACE, uii, "INTERFACE %s\n", name);
    CONFIG_MRTD->interface = interface;
    uii->state = UII_CONFIG_INTERFACE;
    Delete (name);

    LL_Iterate (CONFIG_MRTD->ll_interfaces, ip) {
	if (ip == interface)
	    return (0);
    }

    LL_Add (CONFIG_MRTD->ll_interfaces, interface);
    /* this is a memory to avoid duplication of config_add_module */
    config_add_module (CF_DELIM, "interface", get_config_interface, interface);
    return (1);
}


static void
get_config_interface_gateway (prefix_t * prefix)
{
    config_add_output ("  gateway %s\n", prefix_toa (prefix));
}


static int
config_interface_gateway (uii_connection_t * uii, prefix_t * prefix)
{
    config_notice (TR_TRACE, uii,
		   "gateway %s\n", prefix_toa (prefix));
    add_gateway (prefix, 0, CONFIG_MRTD->interface);
    config_add_module (0, "gateway", get_config_interface_gateway, prefix);
    return (1);
}


#ifdef HAVE_IPV6

static void
get_config_router_ripng ()
{
    int i;
    prefix_t *prefix;
    interface_t *interface;

    config_add_output ("router ripng\n");

    pthread_mutex_lock (&RIPNG->mutex_lock);

    LL_Iterate (RIPNG->ll_networks, prefix) {
	config_add_output ("  network %s\n", prefix_toax (prefix));
    }

    LL_Iterate (RIPNG->ll_interfaces, interface) {
	config_add_output ("  network %s\n", interface->name);
    }

    pthread_mutex_unlock (&RIPNG->mutex_lock);

    pthread_mutex_lock (&MRT->mutex_lock);
    for (i = 0; i <= PROTO_MAX; i++) {
        if (BIT_TEST (MRT->redist[i], (1 << PROTO_RIPNG)))
	    config_add_output ("  redistribute %s\n", proto2string (i));
    }
    pthread_mutex_unlock (&MRT->mutex_lock);
}


/*
 * start RIPNG protocol configuration 
 */
static int
config_router_ripng (uii_connection_t * uii)
{
    if (running && !BIT_TEST (MRT->protocols, 1 << PROTO_RIPNG)) {
        config_notice (TR_ERROR, uii, "You can not do this while running\n");
	return (0);
    }
    CONFIG_MRTD->protocol = PROTO_RIPNG;
    uii->state = UII_CONFIG_ROUTER_RIPNG;
    config_notice (TR_TRACE, uii, "Router RIPng\n");
    if (BIT_TEST (MRT->protocols, 1 << PROTO_RIPNG))
	return (0);

    BIT_SET (MRT->protocols, 1 << PROTO_RIPNG);
    config_add_module (CF_DELIM, "router ripng", get_config_router_ripng, NULL);
    return (1);
}
#endif /* HAVE_IPV6 */


static void
get_config_router_rip ()
{
    int i;
    prefix_t *prefix;
    interface_t *interface;

    config_add_output ("router rip\n");

    pthread_mutex_lock (&RIP->mutex_lock);

    LL_Iterate (RIP->ll_networks, prefix) {
	config_add_output ("  network %s\n", prefix_toax (prefix));
    }

    LL_Iterate (RIP->ll_interfaces, interface) {
	config_add_output ("  network %s\n", interface->name);
    }

    pthread_mutex_unlock (&RIP->mutex_lock);

    pthread_mutex_lock (&MRT->mutex_lock);

    for (i = 0; i <= PROTO_MAX; i++) {
        if (BIT_TEST (MRT->redist[i], (1 << PROTO_RIP)))
	    config_add_output ("  redistribute %s\n", proto2string (i));
    }
    pthread_mutex_unlock (&MRT->mutex_lock);
}


/*
 * start RIP protocol configuration 
 */
static int
config_router_rip (uii_connection_t * uii)
{
    if (running && !BIT_TEST (MRT->protocols, 1 << PROTO_RIP)) {
        config_notice (TR_ERROR, uii, "You can not do this while running\n");
	return (0);
    }

    CONFIG_MRTD->protocol = PROTO_RIP;
    uii->state = UII_CONFIG_ROUTER_RIP;
    config_notice (TR_TRACE, uii, "Router RIP\n");
    if (BIT_TEST (MRT->protocols, 1 << PROTO_RIP))
	return (0);

    BIT_SET (MRT->protocols, 1 << PROTO_RIP);
    config_add_module (CF_DELIM, "router rip", get_config_router_rip, NULL);
    return (1);
}


static void
print_aggregate_address (prefix_t * prefix, u_long option, int view)
{
    char buff[MAXLINE];

    if (prefix->family == AF_INET && view != 0)
	sprintf (buff, " view %d", view);
#ifdef HAVE_IPV6
    else if (prefix->family == AF_INET6 && view != 1)
	sprintf (buff, " view %d", view);
#endif /* HAVE_IPV6 */
    else
	strcpy (buff, "");
    config_add_output ("  aggregate-address %s%s%s%s\n",
		       prefix_toax (prefix),
		       BIT_TEST (option, BGP_AGGOPT_AS_SET) ? " as-set" : "",
	  BIT_TEST (option, BGP_AGGOPT_SUMMARY_ONLY) ? " summary-only" : "",
		       buff);
}


static void
get_config_router_bgp ()
{
    bgp_peer_t *peer;
    int i;
    prefix_t *prefix;

    config_add_output ("router bgp %d\n", BGP->my_as);

    /* I think this is thread-safe... we're just reading data */
    /* NO. someone may change the peer's list, and it may break here */

    pthread_mutex_lock (&BGP->mutex_lock);

    LL_Iterate (BGP->ll_networks, prefix) {
	config_add_output ("  network %s\n", prefix_toax (prefix));
    }

    LL_Iterate (BGP->ll_bgp_peers, peer) {
	char name[MAXLINE];

        pthread_mutex_lock (&peer->mutex_lock);

	strcpy (name, (peer->name)? peer->name: prefix_toa (peer->peer_addr));

	if (peer->name) {
	    config_add_output ("  neighbor %s peer %s src %s iface %s\n", 
		peer->name, prefix_toa (peer->peer_addr), 
		prefix_toa (peer->bind_addr), peer->bind_if->name);
	}

	if (peer->gateway->AS > 0)
	    config_add_output ("  neighbor %s remote-as %d\n", name,
		               peer->gateway->AS);
	if (peer->dlist_in >= 0)
	    config_add_output ("  neighbor %s distribute-list %d in\n",
			name, peer->dlist_in);
	if (peer->dlist_out >= 0)
	    config_add_output ("  neighbor %s distribute-list %d out\n",
		       name, peer->dlist_out);
	if (peer->flist_in >= 0)
	    config_add_output ("  neighbor %s filter-list %d in\n",
			name, peer->flist_in);
	if (peer->flist_out >= 0)
	    config_add_output ("  neighbor %s filter-list %d out\n",
		       name, peer->flist_out);
	if (peer->route_map_in >= 0)
	    config_add_output ("  neighbor %s route-map %d in\n",
			name, peer->route_map_in);
	if (peer->route_map_out >= 0)
	    config_add_output ("  neighbor %s route-map %d out\n",
		       name, peer->route_map_out);

	if (peer->maximum_prefix > 0)
	    config_add_output ("  neighbor %s maximum-prefix %d\n",
		       peer->maximum_prefix);

	if (BIT_TEST (peer->options, BGP_CONNECT_PASSIVE))
	    config_add_output ("  neighbor %s passive\n",
			       name);
	if (BIT_TEST (peer->options, BGP_TRANSPARENT_AS))
	    config_add_output ("  neighbor %s transparent-as\n", name);
	if (BIT_TEST (peer->options, BGP_TRANSPARENT_NEXTHOP))
	    config_add_output ("  neighbor %s transparent-nexthop\n", name);
	if (BIT_TEST (peer->options, BGP_BGP4PLUS_01))
	    config_add_output ("  neighbor %s bgp4+ 1\n", name);
	if (BIT_TEST (peer->options, BGP_BGP4PLUS_AUTO))
	    config_add_output ("  neighbor %s bgp4+ auto\n", name);

	if (peer->HoldTime_Interval >= 0)
	    config_add_output ("  neighbor %s holtime %d\n",
		               name, peer->HoldTime_Interval);
	if (peer->KeepAlive_Interval >= 0)
	    config_add_output ("  neighbor %s keepalive %d\n",
		               name, peer->KeepAlive_Interval);
	if (peer->ConnectRetry_Interval >= 0)
	    config_add_output ("  neighbor %s connectretry %d\n",
		               name, peer->ConnectRetry_Interval);
	if (peer->Start_Interval >= 0)
	    config_add_output ("  neighbor %s starttime %d\n",
		               name, peer->Start_Interval);

        pthread_mutex_unlock (&peer->mutex_lock);
    }
    pthread_mutex_unlock (&BGP->mutex_lock);

    pthread_mutex_lock (&MRT->mutex_lock);
    for (i = 0; i <= PROTO_MAX; i++) {
        if (BIT_TEST (MRT->redist[i], (1 << PROTO_BGP)))
	    config_add_output ("  redistribute %s\n", proto2string (i));
    }
    pthread_mutex_unlock (&MRT->mutex_lock);

    for (i = 0; i < MAX_BGP_VIEWS; i++) {
	view_print_aggregate (i, print_aggregate_address);
    }
}


/* config_router_bgp
 * router bgp %d 
 */
static int
config_router_bgp (uii_connection_t * uii, int as)
{
    if (running && !BIT_TEST (MRT->protocols, 1 << PROTO_BGP)) {
        config_notice (TR_ERROR, uii, "You can not do this while running\n");
	return (0);
    }

    if (as <= 0 || (BGP->my_as != 0 && BGP->my_as != as)) {
	config_notice (TR_TRACE, uii,
		       "BGP is already configured; AS is %d\n", BGP->my_as);
	return (-1);
    }

    CONFIG_MRTD->protocol = PROTO_BGP;
    uii->state = UII_CONFIG_ROUTER_BGP;
    config_notice (TR_TRACE, uii, "Router BGP %d\n", as);

    if (BIT_TEST (MRT->protocols, 1 << PROTO_BGP)) {
	return (0);
    }

    set_BGP (BGP_MY_AS, as, NULL);
    config_add_module (CF_DELIM, "router bgp", get_config_router_bgp, NULL);

    /* probably OK without locking */
    pthread_mutex_lock (&MRT->mutex_lock);
    BIT_SET (MRT->protocols, 1 << PROTO_BGP);
    pthread_mutex_unlock (&MRT->mutex_lock);

    return (1);
}


static void
get_config_router_id (prefix_t *prefix)
{
    config_add_output ("router_id %s\n", prefix_toa (prefix));
}


/*
 * router id %a 
 */
static int
config_router_id (uii_connection_t * uii, prefix_t *prefix)
{
    set_BGP (BGP_MY_ID, prefix_tolong (prefix), NULL);
    config_add_module (0, "router_id", get_config_router_id, prefix);
    trace (TR_TRACE, MRT->trace,
		       "RouterID %s\n", prefix_toa (prefix));
    return (1);
}


static void
get_config_gateway (gateway_t *gateway)
{
    config_add_output ("gateway %s on %s\n", prefix_toa (gateway->prefix),
		       gateway->interface->name);
}


/*
 * router id %a 
 */
static int
config_gateway (uii_connection_t * uii, prefix_t *prefix, char *name)
{
    interface_t *interface;
    gateway_t *gateway;

    interface = find_interface_byname (name);
    if (interface == NULL) {
        config_notice (TR_ERROR, uii,
		       "interface %s not found\n", name);
	Delete (name);
	return (-1);
    }
    config_notice (TR_TRACE, uii, "gateway %s on %s\n", prefix_toa (prefix),
		   interface->name);
    gateway = add_gateway (prefix, 0, interface);
    config_add_module (0, "gateway", get_config_gateway, gateway);
    Delete (name);
    Deref_Prefix (prefix);
    return (1);
}

#define Find_BGP_Peer_ByPrefix(prefix) Find_BGP_Peer(prefix, NULL, NULL)

/* config_router_neighbor_remoteas
 * neighbor %p remote-as %d
 */
static int
config_router_neighbor_remoteas (uii_connection_t * uii,
				 prefix_t * prefix, int as)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	/* I am only one except for automatic addition that adds a peer */
	peer = Add_BGP_Peer (NULL, prefix, NULL, NULL, as, BGP->trace);
	assert (peer);
	if (BGPSIM) {
	    /* BGPSIM should behave as transparent */
            pthread_mutex_lock (&peer->mutex_lock);
	    BIT_SET (peer->options, BGP_TRANSPARENT_AS);
	    BIT_SET (peer->options, BGP_TRANSPARENT_NEXTHOP);
            pthread_mutex_unlock (&peer->mutex_lock);
	}
    }
    else {
	config_notice (TR_ERROR, uii, "Peer is already defined\n");
    	Deref_Prefix (prefix);
    	return (-1);
    }
    if (BGP->my_as == as) {
        pthread_mutex_lock (&peer->mutex_lock);
        pthread_mutex_unlock (&peer->mutex_lock);
	trace (TR_PARSE, BGP->trace, "neighbor %s AS%d as iBGP\n",
	       prefix_toa (prefix), as);
    }
    else {
	trace (TR_PARSE, BGP->trace, "neighbor %s AS%d as eBGP\n",
	       prefix_toa (prefix), as);
    }
    Deref_Prefix (prefix);
    if (running)
	start_bgp_peer (peer);
    return (1);
}


/* no_config_router_neighbor_remoteas
 * no neighbor %p remote-as %d
 */
static int
no_config_router_neighbor_remoteas (uii_connection_t * uii,
				    prefix_t * prefix, int as)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "Peer does not exist\n");
	Deref_Prefix (prefix);
	return (-1);
    }
    /* XXX */

    Destroy_BGP_Peer (peer);
    config_notice (NORM, uii, "Peer deleted\n");
    Deref_Prefix (prefix);
    return (1);
}


static int
config_router_neighbor_n_peer (uii_connection_t * uii, char *name,
			prefix_t * prefix, char *option)
{
    bgp_peer_t *peer;
    prefix_t *usrc = NULL;
    char ifname[MAXLINE];
    interface_t *interface = NULL;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) != NULL) {
	config_notice (TR_ERROR, uii, "Peer %s is already defined\n",
			prefix_toa (prefix));
    	Delete (name);
    	Deref_Prefix (prefix);
    	Delete (option);
    	return (-1);
    }

    if ((peer = Find_BGP_Peer_ByID (name)) != NULL) {
	config_notice (TR_ERROR, uii, "Peer %s is already defined\n", name);
    	Delete (name);
    	Deref_Prefix (prefix);
    	Delete (option);
    	return (-1);
    }

    strcpy (ifname, "");
    if (option) {
        if (parse_line (option, "src %m iface %s", &usrc, ifname) <= 0 &&
            parse_line (option, "iface %s src %m", ifname, &usrc) <= 0) {
    	    Delete (name);
    	    Deref_Prefix (prefix);
    	    Delete (option);
    	    return (-1);
	}
	Delete (option);
    }

    /* XXX usrc should be checked against interface addresses */

    if (ifname[0] && (interface = find_interface_byname (ifname)) == NULL) {
	config_notice (TR_ERROR, uii, "Interface %s is not found\n", ifname);
    	Delete (name);
    	Deref_Prefix (prefix);
    	if (usrc) Deref_Prefix (usrc);
    	return (-1);
    }

    /* I am only one except for automatic addition that adds a peer */
    peer = Add_BGP_Peer (name, prefix, usrc, interface, 0, BGP->trace);
    assert (peer);

	if (BGPSIM) {
	    /* BGPSIM should behave as transparent */
            pthread_mutex_lock (&peer->mutex_lock);
	    BIT_SET (peer->options, BGP_TRANSPARENT_AS);
	    BIT_SET (peer->options, BGP_TRANSPARENT_NEXTHOP);
            pthread_mutex_unlock (&peer->mutex_lock);
	}

    Delete (name);
    Deref_Prefix (prefix);
    if (usrc) Deref_Prefix (usrc);
    return (1);
}



static int
config_router_neighbor_n_remoteas (uii_connection_t * uii, char *name, int as)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "Peer %s is not defined\n", name);
    	Delete (name);
    	return (-1);
    }

    if (BGP->my_as == as) {
        pthread_mutex_lock (&peer->mutex_lock);
        pthread_mutex_unlock (&peer->mutex_lock);
	trace (TR_PARSE, BGP->trace,
	       "neighbor %s AS%d as iBGP\n", name, as);
    }
    else {
	trace (TR_PARSE, BGP->trace,
	       "neighbor %s AS%d as eBGP\n", name, as);
    }
    Delete (name);
    if (running)
	start_bgp_peer (peer);
    return (1);
}


static int
no_config_router_neighbor_n (uii_connection_t * uii, char *name)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "Peer %s does not exist\n", name);
	Delete (name);
	return (-1);
    }
    /* XXX */

    Destroy_BGP_Peer (peer);
    config_notice (TR_INFO, uii, "Peer %s deleted\n", name);
    Delete (name);
    return (1);
}


/* 
 * config_router_neighbor_view
 * just used by BGP to configure peer (at least for now)
 */
static int
config_router_neighbor_view (uii_connection_t * uii, prefix_t * prefix,
			     int view)
{
    bgp_peer_t *peer;
    char str[MAXLINE];

    if (view >= MAX_BGP_VIEWS || view < 0) {
	config_notice (TR_ERROR, uii,
		       "Invalid view %d for BGP neighbor %s [0-%d]\n",
		       view, prefix_toa (prefix), MAX_BGP_VIEWS - 1);
	Deref_Prefix (prefix);
	return (-1);
    }

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	return (-1);
    }

    if (BGP->views[view]) {
	if (BGP->views[view]->family != prefix->family) {
	    config_notice (TR_ERROR, uii, "view %d already defined for %s\n",
			   view, family2string (BGP->views[view]->family));
	    Deref_Prefix (prefix);
	    return (-1);
	}

	trace (TR_PARSE, BGP->trace, "neighbor %s view %d\n",
	       prefix_toa (prefix), view);

        pthread_mutex_lock (&MRT->mutex_lock);
	if (BGP->views[view] == NULL) {
	    /* a new view */
#ifdef HAVE_IPV6
	    if (prefix->family == AF_INET6) {
		BGP->views[view] = New_View (128);
    	        sprintf (str, "BGP4+ view %d", BGP->views[view]->viewno);
	    }
	    else
#endif /* HAVE_IPV6 */
	    {
		BGP->views[view] = New_View (32);
    	        sprintf (str, "BGP view %d", BGP->views[view]->viewno);
	    }

	    BGP->views[view]->viewno = view;
	    BGP->views[view]->trace = trace_copy (BGP->trace);
    	    set_trace (BGP->views[view]->trace, TRACE_PREPEND_STRING, str, 0);

	}
        pthread_mutex_unlock (&MRT->mutex_lock);
	/* locking is not required for an assignment */
        pthread_mutex_lock (&peer->mutex_lock);
	peer->view = view;
        pthread_mutex_unlock (&peer->mutex_lock);
    }
    Deref_Prefix (prefix);
    return (1);
}


static int
config_router_neighbor_n_view (uii_connection_t * uii, char *name,
			     int view)
{
    bgp_peer_t *peer;
    char str[MAXLINE];

    if (view >= MAX_BGP_VIEWS || view < 0) {
	config_notice (TR_ERROR, uii,
		       "Invalid view %d for BGP neighbor %s [0-%d]\n",
		       view, name, MAX_BGP_VIEWS - 1);
	Delete (name);
	return (-1);
    }

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	return (-1);
    }

    if (BGP->views[view]) {
	if (BGP->views[view]->family != peer->peer_addr->family) {
	    config_notice (TR_ERROR, uii, "view %d already defined for %s\n",
			   view, family2string (BGP->views[view]->family));
	    Delete (name);
	    return (-1);
	}

	trace (TR_PARSE, BGP->trace, "neighbor %s view %d\n", name, view);

        pthread_mutex_lock (&MRT->mutex_lock);
	if (BGP->views[view] == NULL) {
	    /* a new view */
#ifdef HAVE_IPV6
	    if (peer->peer_addr->family == AF_INET6) {
		BGP->views[view] = New_View (128);
    	        sprintf (str, "BGP4+ view %d", BGP->views[view]->viewno);
	    }
	    else
#endif /* HAVE_IPV6 */
	    {
		BGP->views[view] = New_View (32);
    	        sprintf (str, "BGP view %d", BGP->views[view]->viewno);
	    }

	    BGP->views[view]->viewno = view;
	    BGP->views[view]->trace = trace_copy (BGP->trace);
    	    set_trace (BGP->views[view]->trace, TRACE_PREPEND_STRING, str, 0);

	}
        pthread_mutex_unlock (&MRT->mutex_lock);
	/* locking is not required for an assignment */
        pthread_mutex_lock (&peer->mutex_lock);
	peer->view = view;
        pthread_mutex_unlock (&peer->mutex_lock);
    }
    Delete (name);
    return (1);
}


/* config_router_neighbor_weight
 */
static int
config_router_neighbor_weight (uii_connection_t * uii,
			       prefix_t * prefix, int weight)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	return (-1);
    }

    /* weight = BGP_PREF; */
    peer->default_pref = weight;

    trace (TR_PARSE, BGP->trace, "neighbor %s weight %d\n",
	   prefix_toa (prefix), weight);
    Deref_Prefix (prefix);
    return (1);
}


static int
config_router_neighbor_n_weight (uii_connection_t * uii,
			       char * name, int weight)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	return (-1);
    }

    /* weight = BGP_PREF; */
    peer->default_pref = weight;

    trace (TR_PARSE, BGP->trace, "neighbor %s weight %d\n", name, weight);
    Delete (name);
    return (1);
}


static int
config_router_neighbor_maximum_prefix (uii_connection_t * uii,
			       prefix_t * prefix, int num)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	return (-1);
    }

    peer->maximum_prefix = num;

    trace (TR_PARSE, BGP->trace, "neighbor %s maximum-prefix %d\n",
	   prefix_toa (prefix), num);
    Deref_Prefix (prefix);
    return (1);
}


static int
config_router_neighbor_n_maximum_prefix (uii_connection_t * uii,
			                 char * name, int num)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	return (-1);
    }

    peer->maximum_prefix = num;

    trace (TR_PARSE, BGP->trace, "neighbor %s maximum-prefix %d\n",
	   name, num);
    Delete (name);
    return (1);
}


/* config_router_neighbor_alias
 */
static int
config_router_neighbor_alias (uii_connection_t * uii,
			      prefix_t * prefix, prefix_t * prefix_alias)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s alias %s\n",
	   prefix_toa (prefix), prefix_toa (prefix_alias));

    if (peer->aliases == NULL)
	peer->aliases = LL_Create (0);

    LL_Add (peer->aliases, prefix_alias);
    Deref_Prefix (prefix);
    return (1);
}


static int
config_router_neighbor_n_alias (uii_connection_t * uii, char *name,
			      prefix_t * prefix_alias)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s alias %s\n",
	   name, prefix_toa (prefix_alias));

    if (peer->aliases == NULL)
	peer->aliases = LL_Create (0);

    LL_Add (peer->aliases, prefix_alias);
    Delete (name);
    return (1);
}


static int
config_router_neighbor_option (uii_connection_t * uii, prefix_t *prefix,
			         char *option)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	Delete (option);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s %s\n",
	   prefix_toa (prefix), option);

    if (strcasecmp (option, "transparent-as") == 0)
        BIT_SET (peer->options, BGP_TRANSPARENT_AS);
    else if (strcasecmp (option, "transparent-nexthop") == 0)
        BIT_SET (peer->options, BGP_TRANSPARENT_NEXTHOP);
    else if (strcasecmp (option, "passive") == 0) {
        BIT_SET (peer->options, BGP_CONNECT_PASSIVE);
    }
    else {
	assert (0);
    }
    Deref_Prefix (prefix);
    Delete (option);
    return (1);
}


static int
config_router_neighbor_n_option (uii_connection_t * uii, char *name,
			         char *option)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	Delete (option);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s %s\n", name, option);

    if (strcasecmp (option, "transparent-as") == 0)
        BIT_SET (peer->options, BGP_TRANSPARENT_AS);
    else if (strcasecmp (option, "transparent-nexthop") == 0)
        BIT_SET (peer->options, BGP_TRANSPARENT_NEXTHOP);
    else if (strcasecmp (option, "passive") == 0) {
        BIT_SET (peer->options, BGP_CONNECT_PASSIVE);
    }
    else {
	assert (0);
    }
    Delete (name);
    Delete (option);
    return (1);
}


static int
config_router_neighbor_list_in_out (uii_connection_t * uii, prefix_t * prefix, 
				    char *op, int num, char *in_or_out)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	Delete (op);
	Delete (in_or_out);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace,
	   "neighbor %s %s %d %s\n", prefix_toa (prefix), op, num, in_or_out);
    if (strcasecmp (op, "distribute-list") == 0) {
	if (strcasecmp (in_or_out, "in") == 0)
            peer->dlist_in = num;
	else
            peer->dlist_out = num;
    }
    else if (strcasecmp (op, "filter-list") == 0) {
	if (strcasecmp (in_or_out, "in") == 0)
    	    peer->flist_in = num;
	else
    	    peer->flist_out = num;
    }
    else {
	assert (0);
    }
    Deref_Prefix (prefix);
    Delete (op);
    Delete (in_or_out);
    return (1);
}


static int
config_router_neighbor_n_list_in_out (uii_connection_t * uii, char *name, 
				      char *op, int num, char *in_or_out)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	Delete (op);
	Delete (in_or_out);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace,
	   "neighbor %s %s %d %s\n", name, op, num, in_or_out);
    if (strcasecmp (op, "distribute-list") == 0) {
	if (strcasecmp (in_or_out, "in") == 0)
            peer->dlist_in = num;
	else
            peer->dlist_out = num;
    }
    else if (strcasecmp (op, "filter-list") == 0) {
	if (strcasecmp (in_or_out, "in") == 0)
    	    peer->flist_in = num;
	else
    	    peer->flist_out = num;
    }
    else {
	assert (0);
    }
    Delete (name);
    Delete (op);
    Delete (in_or_out);
    return (1);
}


static int
config_router_neighbor_routemap_in_out (uii_connection_t * uii,
				  prefix_t * prefix, int num, char *in_or_out)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	Delete (in_or_out);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s route-map %d %s\n",
	   prefix_toa (prefix), num, in_or_out);
    if (strcasecmp (in_or_out, "in") == 0)
        peer->route_map_in = num;
    else
        peer->route_map_out = num;
    Deref_Prefix (prefix);
    Delete (in_or_out);
    return (1);
}


static int
config_router_neighbor_n_routemap_in_out (uii_connection_t * uii,
				       char * name, int num, char *in_or_out)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
        Delete (in_or_out);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s route-map %d out\n",
	   name, num);
    if (strcasecmp (in_or_out, "in") == 0)
        peer->route_map_in = num;
    else
        peer->route_map_out = num;
    Delete (name);
    Delete (in_or_out);
    return (1);
}


static int
config_router_neighbor_bgp4plus (uii_connection_t * uii,
				 prefix_t * prefix, char *version)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
        Delete (version);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s bgp4+ %s\n",
	   prefix_toa (prefix), version);

    BIT_SET (peer->options, BGP_BGP4PLUS);
    if (strcasecmp (version, "1") == 0 || strcasecmp (version, "new") == 0 ||
        strcasecmp (version, "rfc") == 0)
	BIT_SET (peer->options, BGP_BGP4PLUS_01);
    else if (strcasecmp (version, "auto") == 0)
	BIT_SET (peer->options, BGP_BGP4PLUS_AUTO);

    Deref_Prefix (prefix);
    Delete (version);
    return (1);
}


static int
config_router_neighbor_n_bgp4plus (uii_connection_t * uii,
				   char * name, char * version)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
        Delete (version);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s bgp4+ %s\n",
	   name, version);

    BIT_SET (peer->options, BGP_BGP4PLUS);
    if (strcasecmp (version, "1") == 0 || strcasecmp (version, "new") == 0 ||
        strcasecmp (version, "rfc") == 0)
	BIT_SET (peer->options, BGP_BGP4PLUS_01);
    else if (strcasecmp (version, "auto") == 0)
	BIT_SET (peer->options, BGP_BGP4PLUS_AUTO);

    Delete (name);
    Delete (version);
    return (1);
}


static int
config_router_neighbor_time (uii_connection_t * uii,
			     prefix_t * prefix, char *timer, int num)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	Delete (timer);
	return (-1);
    }
    trace (TR_PARSE, BGP->trace, "neighbor %s %s %d\n",
	   prefix_toa (prefix), timer, num);

    if (strcasecmp (timer, "holdtime") == 0) {
        peer->HoldTime_Interval = num;
	Timer_Set_Time (peer->timer_HoldTime, num);
    }
    else if (strcasecmp (timer, "keepalive") == 0) {
        peer->KeepAlive_Interval = num;
	Timer_Set_Time (peer->timer_KeepAlive, num);
    }
    else if (strcasecmp (timer, "connectretry") == 0) {
        peer->ConnectRetry_Interval = num;
	Timer_Set_Time (peer->timer_ConnectRetry, num);
    }
    else if (strcasecmp (timer, "starttime") == 0) {
        peer->Start_Interval = num;
	Timer_Set_Time (peer->timer_Start, num);
    }
    else {
	assert (0);
    }

    Delete (prefix);
    Delete (timer);
    return (1);
}


static int
config_router_neighbor_n_time (uii_connection_t * uii,
			       char * name, char *timer, int num)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s %s %d\n",
	   name, timer, num);

    if (strcasecmp (timer, "holdtime") == 0) {
        peer->HoldTime_Interval = num;
	Timer_Set_Time (peer->timer_HoldTime, num);
    }
    else if (strcasecmp (timer, "keepalive") == 0) {
        peer->KeepAlive_Interval = num;
	Timer_Set_Time (peer->timer_KeepAlive, num);
    }
    else if (strcasecmp (timer, "connectretry") == 0) {
        peer->ConnectRetry_Interval = num;
	Timer_Set_Time (peer->timer_ConnectRetry, num);
    }
    else if (strcasecmp (timer, "starttime") == 0) {
        peer->Start_Interval = num;
	Timer_Set_Time (peer->timer_Start, num);
    }
    else {
	assert (0);
    }

    Delete (name);
    Delete (timer);
    return (1);
}


static int
trace_bgp_neighbor (uii_connection_t * uii, prefix_t *prefix)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	return (-1);
    }

    set_trace (peer->trace, TRACE_ADD_FLAGS, TR_ALL, NULL);
    Deref_Prefix (prefix);
    return (1);
}


static int
trace_bgp_neighbor_n (uii_connection_t * uii, char *name)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	return (-1);
    }

    set_trace (peer->trace, TRACE_ADD_FLAGS, TR_ALL, NULL);
    Delete (name);
    return (1);
}


static int
no_trace_bgp_neighbor (uii_connection_t * uii, prefix_t *prefix)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	return (-1);
    }

    set_trace (peer->trace, TRACE_DEL_FLAGS, TR_ALL, NULL);
    Deref_Prefix (prefix);
    return (1);
}


static int
no_trace_bgp_neighbor_n (uii_connection_t * uii, char *name)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	return (-1);
    }

    set_trace (peer->trace, TRACE_DEL_FLAGS, TR_ALL, NULL);
    Delete (name);
    return (1);
}


static void
get_config_route_map (int num)
{
    char *toa;
    config_add_output ("route-map %d\n", num);
    toa = route_map_toa (num);
    config_add_output ("%s", toa);
    Delete (toa);
}


static int
config_route_map (uii_connection_t * uii, int num)
{
    bgp_attr_t *attr;

    if (num <0 && num >= MAX_ROUTE_MAP) {
	config_notice (TR_ERROR, uii,
	       "wrong route-map number %d (should be 0 <= x < %d)\n",
	       num, MAX_ROUTE_MAP);
	return (-1);
    }
    attr = bgp_new_attr (PROTO_BGP);
    CONFIG_MRTD->route_map = add_route_map (num, attr, 0);
    uii->state = UII_CONFIG_ROUTE_MAP;
    trace (NORM, CONFIG.trace, "CONFIG route-map %d\n", num);
    config_add_module (0, "route-map", get_config_route_map, (void *)num);
    return (1);
}


static int
config_route_map_origin (uii_connection_t * uii, char *str)
{
    int origin, as = 0;

    if (parse_line (str, "igp") >= 1) {
	origin = 0;
    }
    else if (parse_line (str, "egp %d", &as) >= 1) {
	/* as is just for cisco-like syntax */
	origin = 1;
    }
    else if (parse_line (str, "incomplete") >= 1) {
	origin = 2;
    }
    else {
	Delete (str);
	return (-1);
    }

    assert (CONFIG_MRTD->route_map);
    assert (CONFIG_MRTD->route_map->attr);
    CONFIG_MRTD->route_map->attr->origin = origin;
    BGP4_BIT_SET (CONFIG_MRTD->route_map->attr->attribs, PA4_TYPE_ORIGIN);
    Delete (str);
    return (1);
}


static int
config_route_map_weight (uii_connection_t * uii, int weight)
{
    assert (CONFIG_MRTD->route_map);
    assert (CONFIG_MRTD->route_map->attr);
    /* XXX
       CONFIG_MRTD->route_map->attr->pref = weight; */
    return (1);
}


static int
config_route_map_nexthop (uii_connection_t * uii, prefix_t * prefix)
{
    assert (CONFIG_MRTD->route_map);
    assert (CONFIG_MRTD->route_map->attr);

    if (prefix->family == AF_INET) {
	if (BGP4_BIT_TEST (CONFIG_MRTD->route_map->attr->attribs,
			   PA4_TYPE_NEXTHOP)) {
#ifdef HAVE_IPV6
	    if (CONFIG_MRTD->route_map->attr->nexthop->family == AF_INET6) {
		if (CONFIG_MRTD->route_map->attr->nexthop4)
		    Deref_Prefix (CONFIG_MRTD->route_map->attr->nexthop4);
		CONFIG_MRTD->route_map->attr->nexthop4 = prefix;
	    }
	    else
#endif /* HAVE_IPV6 */
	    {
		Deref_Prefix (CONFIG_MRTD->route_map->attr->nexthop);
		CONFIG_MRTD->route_map->attr->nexthop = prefix;
	    }
	}
	else {
	    CONFIG_MRTD->route_map->attr->nexthop = prefix;
	}
	BGP4_BIT_SET (CONFIG_MRTD->route_map->attr->attribs, PA4_TYPE_NEXTHOP);
	return (1);
    }
#ifdef HAVE_IPV6
    else {
	assert (CONFIG_MRTD->route_map);
	assert (CONFIG_MRTD->route_map->attr);

	if (IN6_IS_ADDR_LINKLOCAL (prefix_toaddr6 (prefix))) {
	    if (!BGP4_BIT_TEST (CONFIG_MRTD->route_map->attr->attribs, 
			PA4_TYPE_NEXTHOP))
		CONFIG_MRTD->route_map->attr->nexthop = Ref_Prefix (prefix);
	    if (CONFIG_MRTD->route_map->attr->link_local)
		Deref_Prefix (CONFIG_MRTD->route_map->attr->link_local);
	    CONFIG_MRTD->route_map->attr->link_local = prefix;
	}
	else {
	    if (BGP4_BIT_TEST (CONFIG_MRTD->route_map->attr->attribs,
			       PA4_TYPE_NEXTHOP)) {
		if (CONFIG_MRTD->route_map->attr->nexthop->family == AF_INET) {
		    if (CONFIG_MRTD->route_map->attr->nexthop4)
			Deref_Prefix (CONFIG_MRTD->route_map->attr->nexthop4);
		    CONFIG_MRTD->route_map->attr->nexthop4 =
			CONFIG_MRTD->route_map->attr->nexthop;
		}
		else
		    Deref_Prefix (CONFIG_MRTD->route_map->attr->nexthop);
		CONFIG_MRTD->route_map->attr->nexthop = prefix;
	    }
	    else
		CONFIG_MRTD->route_map->attr->nexthop = prefix;
	}
	BGP4_BIT_SET (CONFIG_MRTD->route_map->attr->attribs, PA4_TYPE_NEXTHOP);
	return (1);
    }
#endif /* HAVE_IPV6 */
    return (-1);
}


static int
config_route_map_metric (uii_connection_t * uii, int metric)
{
    assert (CONFIG_MRTD->route_map);
    assert (CONFIG_MRTD->route_map->attr);
    CONFIG_MRTD->route_map->attr->multiexit = metric;
    BGP4_BIT_SET (CONFIG_MRTD->route_map->attr->attribs, PA4_TYPE_METRIC);
    return (1);
}


static int
config_route_map_localpref (uii_connection_t * uii, int localpref)
{
    assert (CONFIG_MRTD->route_map);
    assert (CONFIG_MRTD->route_map->attr);
    CONFIG_MRTD->route_map->attr->local_pref = localpref;
    BGP4_BIT_SET (CONFIG_MRTD->route_map->attr->attribs, PA4_TYPE_LOCALPREF);
    return (1);
}


static int
config_route_map_community (uii_connection_t * uii, char *string, int opt_num)
{
    int value;

    if (strcasecmp (string, "no-export") == 0)
	value = COMMUNITY_NO_EXPORT;
    else if (strcasecmp (string, "no-advertise") == 0)
	value = COMMUNITY_NO_ADVERTISE;
    else if (isdigit (string[0]))
	value = atoi (string);
    else {
	config_notice (TR_ERROR, uii,
		       "route-map bad community %s\n", string);
	Delete (string);
	return (-1);
    }
    Delete (string);

    assert (CONFIG_MRTD->route_map);
    assert (CONFIG_MRTD->route_map->attr);

    if (!BGP4_BIT_TEST (CONFIG_MRTD->route_map->attr->attribs, 
			PA4_TYPE_COMMUNITY)) {
	CONFIG_MRTD->route_map->attr->community =
	    New_community (1, (u_char *) & value, 0);
    }
    else {
	if (opt_num > 0) {	/* additive */
	    CONFIG_MRTD->route_map->attr->community->len++;
	    CONFIG_MRTD->route_map->attr->community->value =
		(u_long *) ReallocateArray (
			      CONFIG_MRTD->route_map->attr->community->value,
					       u_long,
			       CONFIG_MRTD->route_map->attr->community->len);
	    CONFIG_MRTD->route_map->attr->community->value
		[CONFIG_MRTD->route_map->attr->community->len - 1] = value;
	}
	else {
	    Delete_community (CONFIG_MRTD->route_map->attr->community);
	    CONFIG_MRTD->route_map->attr->community =
		New_community (1, (u_char *) & value, 0);
	}
    }
    BGP4_BIT_SET (CONFIG_MRTD->route_map->attr->attribs, PA4_TYPE_COMMUNITY);
    return (1);
}


static int
config_route_map_aspath (uii_connection_t * uii, char *arg)
{
    char *aspath = arg;

    assert (CONFIG_MRTD->route_map);
    assert (CONFIG_MRTD->route_map->attr);

    if (strncasecmp (aspath, "prepend ", 8) == 0) {
	aspath += 8;
	BIT_SET (CONFIG_MRTD->route_map->flag, ROUTE_MAP_ASPATH_PREPEND);
    }
    if (BGP4_BIT_TEST (CONFIG_MRTD->route_map->attr->attribs, 
			   PA4_TYPE_ASPATH))
	Delete_ASPATH (CONFIG_MRTD->route_map->attr->aspath);
    if ((CONFIG_MRTD->route_map->attr->aspath = 
		aspth_from_string (aspath)) == NULL) {
	config_notice (TR_ERROR, uii,
		       "route-map bad aspath %s\n", aspath);
        Delete (arg);
        return (-1);
    }
    CONFIG_MRTD->route_map->attr->home_AS = 
	bgp_get_home_AS (CONFIG_MRTD->route_map->attr->aspath);
    BGP4_BIT_SET (CONFIG_MRTD->route_map->attr->attribs, PA4_TYPE_ASPATH);
    Delete (arg);
    return (1);
}


static int
config_route_map_atomic (uii_connection_t * uii)
{
    /* Atomic Aggregate */
    assert (CONFIG_MRTD->route_map);
    assert (CONFIG_MRTD->route_map->attr);
    BGP4_BIT_SET (CONFIG_MRTD->route_map->attr->attribs, PA4_TYPE_ATOMICAGG);
    return (1);
}


static int
config_route_map_aggregator (uii_connection_t * uii, int as, prefix_t * prefix)
{
    assert (CONFIG_MRTD->route_map);
    assert (CONFIG_MRTD->route_map->attr);
    if (BGP4_BIT_TEST (CONFIG_MRTD->route_map->attr->attribs, 
		PA4_TYPE_AGGREGATOR))
	Deref_Prefix (CONFIG_MRTD->route_map->attr->aggregator.prefix);
    CONFIG_MRTD->route_map->attr->aggregator.as = as;
    CONFIG_MRTD->route_map->attr->aggregator.prefix = prefix;
    BGP4_BIT_SET (CONFIG_MRTD->route_map->attr->attribs, PA4_TYPE_AGGREGATOR);
    return (1);
}


static int
config_route_map_dpa (uii_connection_t * uii, int as, int value)
{
    /* set dpa */
    assert (CONFIG_MRTD->route_map);
    assert (CONFIG_MRTD->route_map->attr);
    CONFIG_MRTD->route_map->attr->dpa.as = as;
    CONFIG_MRTD->route_map->attr->dpa.value = value;
    BGP4_BIT_SET (CONFIG_MRTD->route_map->attr->attribs, PA4_TYPE_DPA);
    return (1);
}

static void
config_route_map_init ()
{
    set_uii (UII, UII_PROMPT, UII_CONFIG_ROUTE_MAP, "Route Map> ", 0);

    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "route-map %d", 
		      config_route_map, "Defines route map");
    uii_add_command2 (UII_CONFIG_ROUTE_MAP, COMMAND_NORM, "set origin %s",
		      config_route_map_origin, "Sets origin");
    uii_add_command2 (UII_CONFIG_ROUTE_MAP, COMMAND_NORM, "set aspath %S",
		      config_route_map_aspath, "Sets aspath");
    uii_add_command2 (UII_CONFIG_ROUTE_MAP, COMMAND_NORM, "set nexthop %m",
		      config_route_map_nexthop, "Sets nexthop");
    uii_add_command2 (UII_CONFIG_ROUTE_MAP, COMMAND_NORM, "set metric %d",
		      config_route_map_metric, "Sets mertic");
    uii_add_command2 (UII_CONFIG_ROUTE_MAP, COMMAND_NORM, "set weight %d",
		      config_route_map_weight, "Sets weight");
    uii_add_command2 (UII_CONFIG_ROUTE_MAP, COMMAND_NORM, 
		      "set community %s [additive]",
		      config_route_map_community, "Sets community");
    uii_add_command2 (UII_CONFIG_ROUTE_MAP, COMMAND_NORM, "set dpa as %d %d",
		      config_route_map_dpa, "Sets dpa");
    uii_add_command2 (UII_CONFIG_ROUTE_MAP, COMMAND_NORM,
		      "set atomic-aggregate", config_route_map_atomic,
		      "Sets atomic-aggregate");
    uii_add_command2 (UII_CONFIG_ROUTE_MAP, COMMAND_NORM, 
		      "set aggregator as %d %m",
		      config_route_map_aggregator, "Sets aggregator");
    uii_add_command2 (UII_CONFIG_ROUTE_MAP, COMMAND_NORM,
		      "set local-preference %d",
		      config_route_map_localpref, "Sets local preference");
}


#ifdef notdef
static int
config_router_network_weight (uii_connection_t * uii, char *name_or_prefix,
			      int weight)
{
    prefix_t *prefix;
    char name[MAXLINE];
    interface_t *interface;
    LINKED_LIST *ll;

    if (CONFIG_MRTD->protocol != PROTO_RIP && 
	CONFIG_MRTD->protocol != PROTO_RIPNG &&
	CONFIG_MRTD->protocol != PROTO_BGP) {
	Delete (name_or_prefix);
	return (-1);
    }

    if (parse_line (name_or_prefix, "%m", &prefix) >= 1) {

	Delete (name_or_prefix);

	if (CONFIG_MRTD->protocol == PROTO_BGP)
	    LL_Add (BGP->ll_networks, Ref_Prefix (prefix));
	if (CONFIG_MRTD->protocol == PROTO_RIP)
	    LL_Add (RIP->ll_networks, Ref_Prefix (prefix));
#ifdef HAVE_IPV6
	if (CONFIG_MRTD->protocol == PROTO_RIPNG)
	    LL_Add (RIPNG->ll_networks, Ref_Prefix (prefix));
#endif /* HAVE_IPV6 */
	if (CONFIG_MRTD->protocol == PROTO_BGP) {
	    Deref_Prefix (prefix);
	    return (1);
	}

	if ((ll = find_network (prefix)) != NULL) {
	    LL_Iterate (ll, interface) {
		switch (CONFIG_MRTD->protocol) {
		case PROTO_RIP:
		    if (BIT_TEST (RIP->interface_mask, interface->bit))
			break;
		    RIP->interface_mask |= interface->bit;
		    interface->default_pref = weight;
		    config_notice (TR_TRACE, uii,
				   "CONFIG RIP on interface %s weight %d\n",
				   interface->name, weight);
		    break;
#ifdef HAVE_IPV6
		case PROTO_RIPNG:
		    if (BIT_TEST (RIPNG->interface_mask, interface->bit))
			break;
		    if (!BIT_TEST (interface->flags, IFF_MULTICAST) &&
			BIT_TEST (interface->flags, IFF_POINTOPOINT) &&
			(interface->link_local == NULL ||
			 interface->link_local->broadcast == NULL)) {
			config_notice (TR_ERROR, uii,
				     "RIPNG on interface %s ignored "
			    "due to no multicast or link-local dest addr\n",
				       interface->name);
			break;
		    }
		    RIPNG->interface_mask |= interface->bit;
		    interface->default_pref = weight;
		    config_notice (TR_TRACE, uii,
				 "CONFIG RIPNG on interface %s weight %d\n",
				   interface->name, weight);
		    break;
#endif /* HAVE_IPV6 */
		default:
		    assert (0);
		    break;
		}
	    }
	    LL_Destroy (ll);
	    Deref_Prefix (prefix);
	    return (1);
	}
	/* else {
	    Deref_Prefix (prefix);
	    return (-1);
	} */
    }
    else if (parse_line (name_or_prefix, "%s", name) >= 1) {

	Delete (name_or_prefix);
/*
 * This part must be the last otherwise some ipv6 address will be mistook 
 */
	if ((interface = find_interface_byname (name)) == NULL) {
	    config_notice (TR_ERROR, uii,
		   "could not find interface %s\n", name);
	    return (-1);
	}

	if (CONFIG_MRTD->protocol == PROTO_RIP) {
	    RIP->interface_mask |= interface->bit;
	    interface->default_pref = weight;
	    config_notice (TR_TRACE, uii,
			   "CONFIG RIP on interface %s\n", interface->name);
	    LL_Add (RIP->ll_networks, Ref_Prefix (interface->primary->prefix));
	    return (1);
	}
#ifdef HAVE_IPV6
	else if (CONFIG_MRTD->protocol == PROTO_RIPNG) {
	    if (!BIT_TEST (RIPNG->interface_mask, interface->bit)) {
		if (!BIT_TEST (interface->flags, IFF_MULTICAST) &&
		    BIT_TEST (interface->flags, IFF_POINTOPOINT) &&
		    (interface->link_local == NULL ||
		     interface->link_local->broadcast == NULL)) {
		    config_notice (TR_ERROR, uii,
				   "RIPNG on interface %s ignored "
			    "due to no multicast or link-local dest addr\n",
				   interface->name);
		}
		else {
		    RIPNG->interface_mask |= interface->bit;
		    interface->default_pref = weight;
		    config_notice (TR_TRACE, uii,
				   "CONFIG RIPNG on interface %s\n",
				   interface->name);
	            LL_Add (RIPNG->ll_networks, Ref_Prefix (interface->primary6->prefix));
		}
	    }
	    return (1);
	}
#endif /* HAVE_IPV6 */
	else if (CONFIG_MRTD->protocol == PROTO_BGP) {
	    /* not supported */
	    return (-1);
	}
	else {
	    /* should not happen */
	    assert (0);
	}
    }
    else {
	Delete (name_or_prefix);
	return (-1);
    }
    assert (0);
    /* NOT REACHED */
    return (-1); /* shut up the compiler */
}
#endif


static int
config_router_network_prefix (uii_connection_t * uii, prefix_t *prefix)
{
    interface_t *interface;
    LINKED_LIST *ll;

    if (CONFIG_MRTD->protocol != PROTO_RIP && 
	CONFIG_MRTD->protocol != PROTO_RIPNG &&
	CONFIG_MRTD->protocol != PROTO_BGP) {
	Deref_Prefix (prefix);
	return (-1);
    }

    if (CONFIG_MRTD->protocol == PROTO_BGP) {
	LL_Add (BGP->ll_networks, Ref_Prefix (prefix));
	Deref_Prefix (prefix);
	return (1);
    }
    else if (CONFIG_MRTD->protocol == PROTO_RIP)
	LL_Add (RIP->ll_networks, Ref_Prefix (prefix));
#ifdef HAVE_IPV6
    else /* RIPNG */
	LL_Add (RIPNG->ll_networks, Ref_Prefix (prefix));
#endif /* HAVE_IPV6 */

    if ((ll = find_network (prefix)) != NULL) {
	LL_Iterate (ll, interface) {
	    if (CONFIG_MRTD->protocol == PROTO_RIP) {
		if (!BIT_TEST (RIP->interface_mask, interface->bit)) {
		    RIP->interface_mask |= interface->bit;
		    interface->default_pref = RIP_PREF;
		    config_notice (TR_TRACE, uii, "RIP on interface %s\n",
				   interface->name);
		}
	    }
#ifdef HAVE_IPV6
	    else { /* RIPNG */
		if (!BIT_TEST (RIPNG->interface_mask, interface->bit)) {
		    if (!BIT_TEST (interface->flags, IFF_MULTICAST) &&
			BIT_TEST (interface->flags, IFF_POINTOPOINT) &&
			(interface->link_local == NULL ||
			 interface->link_local->broadcast == NULL)) {
			config_notice (TR_ERROR, uii,
				     "RIPNG on interface %s ignored "
			    "due to no multicast or link-local dest addr\n",
				       interface->name);
		    }
		    else {
		        RIPNG->interface_mask |= interface->bit;
		        interface->default_pref = RIPNG_PREF;
		        config_notice (TR_TRACE, uii,
				       "RIPNG on interface %s\n",
				       interface->name);
		    }
		}
	    }
#endif /* HAVE_IPV6 */
	}
	LL_Destroy (ll);
	Deref_Prefix (prefix);
	return (1);
    }
    /* else {
	Deref_Prefix (prefix);
	return (-1);
    } */
    Deref_Prefix (prefix);
    return (1);
}



static int
config_router_network_interface (uii_connection_t * uii, char *name)
{
    interface_t *interface;

    if (CONFIG_MRTD->protocol != PROTO_RIP && 
	CONFIG_MRTD->protocol != PROTO_RIPNG) {
	Delete (name);
	return (-1);
    }
    if ((interface = find_interface_byname (name)) == NULL) {
	config_notice (TR_ERROR, uii,
		       "could not find interface %s\n", name);
	Delete (name);
	return (-1);
    }
    Delete (name);

	if (CONFIG_MRTD->protocol == PROTO_RIP) {
	    RIP->interface_mask |= interface->bit;
	    interface->default_pref = RIP_PREF;
	    config_notice (TR_TRACE, uii,
			   "RIP on interface %s\n", interface->name);
	    LL_Add (RIP->ll_interfaces, interface);
	    return (1);
	}
#ifdef HAVE_IPV6
	else { /* RIPNG */
	    if (!BIT_TEST (RIPNG->interface_mask, interface->bit)) {
		if (!BIT_TEST (interface->flags, IFF_MULTICAST) &&
		    BIT_TEST (interface->flags, IFF_POINTOPOINT) &&
		    (interface->link_local == NULL ||
		     interface->link_local->broadcast == NULL)) {
		    config_notice (TR_ERROR, uii,
				   "RIPNG on interface %s ignored "
			    "due to no multicast or link-local dest addr\n",
				   interface->name);
		}
		else {
		    RIPNG->interface_mask |= interface->bit;
		    interface->default_pref = RIPNG_PREF;
		    config_notice (TR_TRACE, uii,
				   "RIPNG on interface %s\n",
				   interface->name);
	    	    LL_Add (RIPNG->ll_interfaces, interface);
		}
	        return (1);
	    }
	    /* already set but don't care */
	    return (0);
	}
#endif /* HAVE_IPV6 */
    return (-1); /* shut up the compiler */
}


static int
config_router_aggregate (uii_connection_t * uii, prefix_t * prefix, 
			 char *rest)
{
    u_long option;
    char opt[4][MAXLINE];
    int view = 0, n;

#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6)
	view = 1;
#endif /* HAVE_IPV6 */

    opt[0][0] = opt[1][0] = opt[2][0] = opt[3][0] = '\0';
    option = 0;
    if (rest && (n = parse_line (rest, "%s %s %s %s",
				 opt[0], opt[1], opt[2], opt[3])) >= 1) {
	int i;

	for (i = 0; i < n; i++) {

	    if (strcasecmp (opt[i], "as-set") == 0) {
	        BIT_SET (option, BGP_AGGOPT_AS_SET);
	    }
	    else if (strcasecmp (opt[i], "summary-only") == 0) {
	        BIT_SET (option, BGP_AGGOPT_SUMMARY_ONLY);
	    }
	    else if (i < n - 1 && strcasecmp (opt[i], "view") == 0
	        && isdigit (opt[i + 1][0])) {
	        view = atoi (opt[i + 1]);
		i++;
	    }
	}
    }

    if (rest)
	Delete (rest);

    if (BGP->views[view] == NULL) {
	config_notice (TR_TRACE, uii,
		    "CONFIG aggregate-address view %d not defined\n", view);
	Deref_Prefix (prefix);
	return (-1);
    }
    if (BGP->views[view]->family != prefix->family) {
	config_notice (TR_TRACE, uii,
		"CONFIG aggregate-address view %d family %s not %s\n", view,
		       family2string (BGP->views[view]->family),
		       family2string (prefix->family));
	Deref_Prefix (prefix);
	return (-1);
    }
    view_add_aggregate (BGP->views[view], prefix, option);
    config_notice (TR_TRACE, uii,
		   "CONFIG aggregate-address %s%s%s\n",
		   prefix_toax (prefix),
		   BIT_TEST (option, BGP_AGGOPT_AS_SET) ? " as-set" : "",
	 BIT_TEST (option, BGP_AGGOPT_SUMMARY_ONLY) ? " summary-only" : "");
    Deref_Prefix (prefix);
    return (1);
}


static int
no_config_router_aggregate (uii_connection_t * uii, prefix_t * prefix, 
			    char *rest)
{
    int view = 0;

#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6)
	view = 1;
#endif
    if (rest) {
	parse_line (rest, "view %d", &view);
	Delete (rest);
    }

    if (BGP->views[view] == NULL) {
	config_notice (TR_ERROR, uii,
		 "no aggregate-address view %d not defined\n", view);
	Deref_Prefix (prefix);
	return (-1);
    }
    if (BGP->views[view]->family != prefix->family) {
	config_notice (TR_ERROR, uii,
		   "no aggregate-address view %d family %s not %s\n",
		       view,
		       family2string (BGP->views[view]->family),
		       family2string (prefix->family));
	Deref_Prefix (prefix);
	return (-1);
    }
    view_del_aggregate (BGP->views[view], prefix);
    config_notice (TR_TRACE, uii,
		   "CONFIG no aggregate-address %s view %d\n",
		   prefix_toa (prefix), view);
    Deref_Prefix (prefix);
    return (1);
}



/* config_redistribute
 * Redistribute a given protocol into another protocol 
 */
static int
config_redistribute (uii_connection_t * uii, char *proto_string)
{
    u_int proto;

    if (!strcasecmp (proto_string, "kernel")) {
	proto = PROTO_KERNEL;
    }
    else if (!strcasecmp (proto_string, "static")) {
	proto = PROTO_STATIC;
    }
    else if (!strcasecmp (proto_string, "direct")) {
	/* for compatibility */
	proto = PROTO_CONNECTED;
    }
    else if (!strcasecmp (proto_string, "connected")) {
	proto = PROTO_CONNECTED;
    }
    else if (!strcasecmp (proto_string, "rip")) {
	proto = PROTO_RIP;
    }
    else if (!strcasecmp (proto_string, "ospf")) {
	proto = PROTO_OSPF;
    }
    else if (!strcasecmp (proto_string, "bgp")) {
	proto = PROTO_BGP;
    }
#ifdef HAVE_IPV6
    else if (!strcasecmp (proto_string, "ripng")) {
	proto = PROTO_RIPNG;
    }
#endif
    else {
	Delete (proto_string);
	return (-1);
    }

    if (CONFIG_MRTD->protocol == proto) {
	config_notice (TR_ERROR, uii,
		 "%s redistribute %s -- the same protocols!\n",
		       proto2string (CONFIG_MRTD->protocol), proto2string (proto));
	Delete (proto_string);
	return (-1);
    }

    MRT->redist[proto] |= (1 << CONFIG_MRTD->protocol);
    config_notice (TR_TRACE, uii,
		   "CONFIG %s redistribute %s\n",
		   proto2string (CONFIG_MRTD->protocol), proto2string (proto));
    Delete (proto_string);
    return (1);
}


/* config_redistribute
 * Redistribute a given protocol into another protocol 
 * no redistribute (bgp|ripng|static|direct)
 */
static int
no_config_redistribute (uii_connection_t * uii, char *proto_string)
{
    u_int proto;

    if (!strcasecmp (proto_string, "kernel")) {
	proto = PROTO_KERNEL;
    }
    else if (!strcasecmp (proto_string, "static")) {
	proto = PROTO_STATIC;
    }
    else if (!strcasecmp (proto_string, "direct")) {
	/* for compatibility */
	proto = PROTO_CONNECTED;
    }
    else if (!strcasecmp (proto_string, "connected")) {
	proto = PROTO_CONNECTED;
    }
    else if (!strcasecmp (proto_string, "rip")) {
	proto = PROTO_RIP;
    }
    else if (!strcasecmp (proto_string, "ospf")) {
	proto = PROTO_OSPF;
    }
    else if (!strcasecmp (proto_string, "bgp")) {
	proto = PROTO_BGP;
    }
#ifdef HAVE_IPV6
    else if (!strcasecmp (proto_string, "ripng")) {
	proto = PROTO_RIPNG;
    }
#endif
    else {
	Delete (proto_string);
	return (-1);
    }

    MRT->redist[proto] &= ~(1 << CONFIG_MRTD->protocol);
    config_notice (TR_TRACE, uii,
		   "CONFIG no %s redistribute %s\n",
		   proto2string (CONFIG_MRTD->protocol), proto2string (proto));
    Delete (proto_string);
    return (1);
}


static int
config_dump_bgp_common (uii_connection_t * uii, char *filename,
			char *unit, time_t * time)
{
    if (unit[0] && isdigit (unit[0])) {
	int i;
	u_int u;
	u = atoi (unit);
	for (i = 0; i < sizeof (unit) && unit[i] && isdigit (unit[i]); i++);
	if (i < sizeof (unit) && isalpha (unit[i])) {
	    if (strncasecmp (unit + i, "m", 1) == 0)
		u *= 60;
	    else if (strncasecmp (unit + i, "h", 1) == 0)
		u *= (60 * 60);
	}
	*time = u;
    }
    return (1);
}


static char *
time_to_unit (time_t t)
{
    char *buff;

    THREAD_SPECIFIC_STRAGE (buff);
    if ((t % 3600) == 0)
	sprintf (buff, "%ldh", t / 3600);
    else if ((t % 60) == 0)
	sprintf (buff, "%ldm", t / 60);
    else
	sprintf (buff, "%ld", t);
    return (buff);

}

static void
get_config_dump_bgp_updates ()
{
    config_add_output ("dump bgp updates %s %s\n", BGP->dump_update_form,
		       time_to_unit (BGP->dump_update_interval));
}


static void
get_config_dump_bgp_view (int id)
{
    config_add_output ("dump bgp view %d %s %s\n", id,
			   BGP->dump_route_form[id],
			   time_to_unit (BGP->dump_route_interval[id]));
}


static int
config_dump_bgp_updates (uii_connection_t * uii, char *filename, char *unit)
{
    time_t t = 0;

    if (config_dump_bgp_common (uii, filename, unit, &t) >= 0 && filename[0]) {
	set_BGP (BGP_DUMP_UPDATE_FORM, filename, t, 0);
	config_notice (TR_TRACE, uii,
		       "CONFIG dump bgp updates %s %d\n", filename, t);
	config_add_module (0, "bgp dump", get_config_dump_bgp_updates, NULL);
	Delete (filename);
	Delete (unit);
	return (1);
    }
    Delete (filename);
    Delete (unit);
    return (-1);
}


static int
config_dump_bgp_view (uii_connection_t * uii,
		      int id, char *filename, char *unit)
{
    time_t t = 0;

    if (config_dump_bgp_common (uii, filename, unit, &t) >= 0 && filename[0]) {
	set_BGP (BGP_DUMP_ROUTE_FORM, id, filename, t, 0);
	config_notice (TR_TRACE, uii,
		       "CONFIG dump bgp view %d %s %d\n", id, filename, t);
	config_add_module (0, "bgp dump", get_config_dump_bgp_view, id);
	Delete (filename);
	Delete (unit);
	return (1);
    }
    Delete (filename);
    Delete (unit);
    return (-1);
}


static void
get_config_dump_bgp_routes ()
{
    config_add_output ("dump bgp routes %s %s\n", BGP->dump_route_form[0],
		       time_to_unit (BGP->dump_route_interval[0]));
}



static int
config_dump_bgp_routes (uii_connection_t * uii, char *filename, char *unit)
{
    time_t t = 0;
    if (config_dump_bgp_common (uii, filename, unit, &t) >= 0 && filename[0]) {
	set_BGP (BGP_DUMP_ROUTE_FORM, 0, filename, t, 0);
#ifdef HAVE_IPV6
	set_BGP (BGP_DUMP_ROUTE_FORM, 1, filename, t, 0);
#endif /* HAVE_IPV6 */
	config_notice (TR_TRACE, uii,
		       "CONFIG dump bgp routes %s %d\n", filename, t);
	config_add_module (0, "bgp dump", get_config_dump_bgp_routes, NULL);
	Delete (filename);
	Delete (unit);
	return (1);
    }
    Delete (filename);
    Delete (unit);
    return (-1);
}


static int
kill_peer (uii_connection_t * uii, char *whom)
{
    prefix_t *prefix;
    bgp_peer_t *peer;

    if (strcmp (whom, "*") == 0) {
	bgp_kill_all ();
	return (1);
    }
    else if (parse_line (whom, "%m", &prefix) >= 1) {

	if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	    config_notice (NORM, uii, "No peer %s\n", prefix_toa (prefix));
	    return (-1);
	}

	return (bgp_kill_peer2 (peer));
    }
    return (-1);
}


static int
config_mrt_reboot (uii_connection_t *uii)
{
#if defined (__linux__) && defined (HAVE_LIBPTHREAD)
    config_notice (TR_ERROR, uii,
        "This command is unavailable on Linux with Pthread\n");
    return (0);
#else
    mrt_reboot ();
    /* never returns */
    return (1);
#endif
}


int
init_mrtd_config (trace_t * trace)
{
    CONFIG_MRTD = New (config_mrtd_t);
    CONFIG_MRTD->ll_static_routes = NULL;
    CONFIG_MRTD->interface = NULL;
    CONFIG_MRTD->ll_interfaces = LL_Create (0);
    CONFIG_MRTD->ll_static_routes = LL_Create (0);
    CONFIG_MRTD->password = NULL;

    UII->initial_state = 0;	/* password required */

    set_uii (UII, UII_PROMPT, UII_UNPREV, "Password> ", 0);
    set_uii (UII, UII_PROMPT, UII_NORMAL, "MRTd> ", 0);
    set_uii (UII, UII_PROMPT, UII_CONFIG, "Config> ", 0);
    set_uii (UII, UII_PROMPT, UII_CONFIG_ROUTER_BGP, "Router BGP> ", 0);
    set_uii (UII, UII_PROMPT, UII_CONFIG_ROUTER_RIP, "Router RIP> ", 0);
#ifdef HAVE_IPV6
    set_uii (UII, UII_PROMPT, UII_CONFIG_ROUTER_RIPNG, "Router RIPng> ", 0);
#endif /* HAVE_IPV6 */
    set_uii (UII, UII_PROMPT, UII_CONFIG_INTERFACE, "Interface> ", 0);

    /* Password */
    uii_add_command2 (UII_UNPREV, COMMAND_MATCH_FIRST, "", uii_check_passwd,
		      "Password");

    /* Interactive commands */
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show rip", show_rip,
		      "Show RIP status");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show rip routes",
		      show_rip_routing_table, "Show RIP routing table");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show bgp", show_bgp,
		      "Show BGP peers and their status");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show view %d", show_view,
		      "Show the BGP routing table for the view");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show bgp routes",
		      show_bgp_routing_table, "Show BGP routing table");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show bgp regexp %S",
		      show_bgp_rt_regexp,
		      "Show BGP routes matching the regular expression");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "show bgp neighbors (%m|%n|*) errors",
		      show_bgp_neighbors_errors,
		      "Show BGP recent errors [with the peer]");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "show bgp neighbors (%m|%n|*) routes",
		      show_bgp_neighbors_routes,
		      "Show BGP routes sent [to the peer]");

    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show ip", show_ip_routes,
		      "Show the central routing table");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show rib routes", 
		      show_ip_routes, "Show the central routing table");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show rib", show_rib_status,
		      "Show the central routing status");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show interfaces",
		      show_interfaces, "Show all interfaces available");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "clear bgp %s", kill_peer,
		      "Close BGP peering session");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "config",
		      start_config, "Configure MRTd");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show config", show_config,
		      "Display current configuration");
/*
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "write", config_write,
		      "Save configuration file to disk");
*/
    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "write", config_write,
		      "Save configuration file to disk");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "reboot", config_mrt_reboot,
		      "Reboot MRTd");
#ifdef HAVE_IPV6
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show ripng", ripng_show,
		      "Show RIPng Status");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show ripng routes",
		      ripng_show_routing_table,
		      "Show RIPng routing table");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show ipv6", show_ipv6_routes,
		      "Show IPv6 RIB routing table");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show rib6 routes", 
		      show_ipv6_routes, "Show IPv6 RIB routing table");
#endif /* HAVE_IPV6 */

    /* Config mode */

    uii_add_command2 (UII_CONFIG_ALL, COMMAND_NORM, "show config", show_config,
		      "Display current configuration");

    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "router rip",
		      config_router_rip, "Enables RIP routing protocol");
#ifdef HAVE_IPV6
    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "router ripng",
		      config_router_ripng, "Enables RIPng routing protocol");
#endif /* HAVE_IPV6 */
    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "router bgp %d",
		      config_router_bgp,
		      "Enables BGP (BGP4+) routing protocol with AS number");

    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "route %m %m %S",
		      config_ip_route, "Add a static route");
    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "ip route %m %m %S",
		      config_ip_route, "Obsolete");
    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "no route %m %m",
		      no_config_ip_route, "Delete static route");

    uii_add_command2 (UII_CONFIG, COMMAND_NORM,
		      "as-path access-list %d (permit|deny) %S",
		      config_ip_as_filter, "Defines AS path filter");

    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "interface", config_interface,
		      "Define interface");
    uii_add_command2 (UII_CONFIG_INTERFACE, COMMAND_NORM, "gateway %m",
		      config_interface_gateway, 
		      "Define gateways on the interface");

    /* move to uii_normal */
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "trace bgp neighbor %m", trace_bgp_neighbor,
		      "Enable trace bgp neighbor");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "no trace bgp neighbor %m", 
		      no_trace_bgp_neighbor, "Disable trace bgp neighbor");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "trace bgp neighbor %n", trace_bgp_neighbor_n,
		      "Enable trace bgp neighbor");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "no trace bgp neighbor %n", 
		      no_trace_bgp_neighbor_n, "Disable trace bgp neighbor");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "trace bgp view (*|inet|inet6|%d)", 
		      trace_bgp_view, "Enable trace bgp view");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "no trace bgp view (*|inet|inet6|%d)", 
		      no_trace_bgp_view, "Disable trace bgp view");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "trace rib (*|inet|inet6)", 
		      trace_rib, "Enable trace rib");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "no trace rib (*|inet|inet6)", 
		      no_trace_rib, "Enable trace rib");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "trace bgp", trace_bgp,
		      "Enable trace BGP");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "no trace bgp", no_trace_bgp,
		      "Disable trace BGP");

    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "dump bgp view %d %s %s",
	      config_dump_bgp_view, "Dumps BGP routing table for the view");
    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "dump bgp routes %s %s",
		      config_dump_bgp_routes, "Dumps BGP routing table");
    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "dump bgp updates %s %s",
	    config_dump_bgp_updates, "Dumps BGP updates and state changes");

    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "router_id %a",
	    	      config_router_id, "Set router id");
    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "gateway %M on %n",
	    	      config_gateway, "Define gateway on interface");

    /* Router [rip|bgp|ripng] */

    uii_add_command2 (UII_CONFIG_ROUTER_RIP, COMMAND_NORM, "network %p",
		      config_router_network_prefix, 
		      "Turns ON RIP routing for the network");
    uii_add_command2 (UII_CONFIG_ROUTER_RIP, COMMAND_NORM, "network %n",
		      config_router_network_interface, 
		      "Turns ON RIP routing on the interface");
#ifdef HAVE_IPV6
    uii_add_command2 (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM, "network %P",
		      config_router_network_prefix, 
		      "Turns ON RIPng routing for the network");
    uii_add_command2 (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM, "network %n",
		      config_router_network_interface, 
		      "Turns ON RIPng routing on the interface");
#endif /* HAVE_IPV6 */
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM, "network %m",
		      config_router_network_prefix, 
		      "Originates routes in BGP");
/*
    uii_add_command2 (UII_CONFIG_ROUTER, COMMAND_NORM,
		  "network (%s|%m) weight %d", config_router_network_weight,
		      "Turns ON routing on the network with the weight");
*/

    uii_add_command2 (UII_CONFIG_ROUTER_RIP, COMMAND_NORM,
	"redistribute (static|ospf|bgp|direct|connected|kernel)",
		      config_redistribute,
		      "Redistributes route from the protocol");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
	"no redistribute (static|ospf|bgp|direct|connected|kernel)",
		      no_config_redistribute,
		      "Not redistribute route from the protocol");
#ifdef HAVE_IPV6
    uii_add_command2 (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM,
	"redistribute (static|ospf|bgp|direct|connected|kernel)",
		      config_redistribute,
		      "Redistributes route from the protocol");
    uii_add_command2 (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM,
	"no redistribute (static|ospf|bgp|direct|connected|kernel)",
		      no_config_redistribute,
		      "Not redistribute route from the protocol");
#endif /* HAVE_IPV6 */
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
	"redistribute (static|ospf|rip|ripng|direct|connected|kernel)",
		      config_redistribute,
		      "Redistribute route from the protocol");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
	"no redistribute (static|ospf|rip|ripng|direct|connected|kernel)",
		      no_config_redistribute,
		      "Not redistribute route from the protocol");

    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m remote-as %d",
		   config_router_neighbor_remoteas, "Adds an BGP neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "no neighbor %m remote-as %d",
	     no_config_router_neighbor_remoteas, "Deletes an BGP neighbor");

    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m view %d", config_router_neighbor_view,
		      "Assigns a view for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m weight %d", config_router_neighbor_weight,
		      "Sets a weight for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m maximum-prefix %d", 
		       config_router_neighbor_maximum_prefix,
		      "Sets max number of prefixes for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m alias %n",
		      config_router_neighbor_alias, "Obsolete");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		  "neighbor %m (transparent-as|transparent-nexthop|passive)",
		      config_router_neighbor_option,
		      "Sets an option for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
	"neighbor %m (distribute-list|filter-list) %d (in|out)",
		      config_router_neighbor_list_in_out,
		      "Applies the filter for incoming or outgoing");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m route-map %d (in|out)",
		      config_router_neighbor_routemap_in_out,
		      "Applies the route map for incoming or outgoing");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m bgp4+ (0|1|old|new|rfc|auto)",
		      config_router_neighbor_bgp4plus,
		      "Specifies BGP4+ packet format");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		"neighbor %m (holdtime|keepalive|connectretry|starttime) %d",
		      config_router_neighbor_time,
		      "Sets timer for neighbor");

    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n peer %m %S",
		   config_router_neighbor_n_peer, "Adds an BGP neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "no neighbor %n",
	     no_config_router_neighbor_n, "Deletes an BGP neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n remote-as %d",
		       config_router_neighbor_n_remoteas, 
		      "Sets AS number to the neighbor");

    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n view %d", config_router_neighbor_n_view,
		      "Assigns view for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n weight %d", config_router_neighbor_n_weight,
		      "Sets weight for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n maximum-refix %d", 
		       config_router_neighbor_n_maximum_prefix,
		      "Sets max number of prefixes for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n alias %n",
		      config_router_neighbor_n_alias, "Obsolete");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		  "neighbor %n (transparent-as|transparent-nexthop|passive)",
		      config_router_neighbor_n_option,
		      "Sets an option for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
	"neighbor %n (distribute-list|filter-list) %d (in|out)",
		      config_router_neighbor_n_list_in_out,
		      "Applies the filter for incoming or outgoing");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n route-map %d (in|out)",
		      config_router_neighbor_n_routemap_in_out,
		      "Applies the route map for incoming or outgoing");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n bgp4+ (0|1|old|new|rfc|auto)",
		      config_router_neighbor_n_bgp4plus,
		      "Specifies BGP4+ packet format");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		"neighbor %n (holdtime|keepalive|connectretry|starttime) %d",
		      config_router_neighbor_n_time,
		      "Sets timer for neighbor");

    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "aggregate-address %m %S", config_router_aggregate,
		      "Creates an aggregate entry to prefix");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "no aggregate-address %m %S", 
		       no_config_router_aggregate,
		      "Deletes an aggregate entry to prefix");

#ifdef HAVE_IPV6
    uii_add_command2 (UII_CONFIG_ROUTER_RIP, COMMAND_NORM,
		      "distribute-list %d (in|out) [%s]",
		      config_distribute_list,
		      "Applies access list to route [on interface]");
    uii_add_command2 (UII_CONFIG_ROUTER_RIPNG, COMMAND_NORM,
		      "distribute-list %d (in|out) [%s]",
		      config_distribute_list,
		      "Applies access list to route [on interface]");
#endif /* HAVE_IPV6 */

    config_route_map_init ();

#ifdef notdef
    MRT->redist[PROTO_STATIC] |= (1 << PROTO_RIP);
    MRT->redist[PROTO_CONNECTED] |= (1 << PROTO_RIP);
/*  MRT->redist[PROTO_KERNEL] |= (1 << PROTO_RIP); */

    MRT->redist[PROTO_STATIC] |= (1 << PROTO_BGP);
    MRT->redist[PROTO_CONNECTED] |= (1 << PROTO_BGP);
/*  MRT->redist[PROTO_KERNEL] |= (1 << PROTO_BGP); */

#ifdef HAVE_IPV6
    MRT->redist[PROTO_STATIC] |= (1 << PROTO_RIPNG);
    MRT->redist[PROTO_CONNECTED] |= (1 << PROTO_RIPNG);
/*  MRT->redist[PROTO_KERNEL] |= (1 << PROTO_RIPNG); */
#endif
#endif

    return (1);
}


void 
config_create_default ()
{
    char *tmp;
    CONFIG.ll_modules = LL_Create (0);

    tmp = strdup (
    "#####################################################################");
    config_add_module (0, "comment", get_comment_config, tmp);

    tmp = malloc (512);
    sprintf (tmp, "# MRTd -- MRT version %s ", MRT_VERSION);
    config_add_module (0, "comment", get_comment_config, tmp);

    tmp = strdup (
    "#####################################################################");
    config_add_module (0, "comment", get_comment_config, tmp);

    config_add_module (0, "comment", get_comment_config, strdup ("#"));
    config_add_module (0, "debug", get_debug_config, NULL);
    config_add_module (0, "comment", get_comment_config, strdup ("#"));

    set_uii (UII, UII_PASSWORD, strdup ("foo"), 0);
    config_add_module (0, "password", get_uii_password, NULL);
}
